# Classes to represent metadata for command line tools.
from pathlib import Path
from collections import OrderedDict
from abc import abstractmethod
import re
import uuid
from ruamel.yaml import safe_load
from capanno_utils.repo_config import *
from capanno_utils.exceptions import InIndexError, NotInIndexError
from ...helpers.get_paths import *
from ...classes.metadata.metadata_base import MetadataBase
from ...classes.metadata.shared_properties import CodeRepository, Person, WebSite, Keyword, IOObjectItem, IOArrayItem
from ...helpers.get_metadata_from_biotools import make_tool_metadata_kwargs_from_biotools
from ...classes.metadata.common_functions import _mk_hashes, CommonPropsMixin, WorkflowLanguageStatusMixin, SoftwarePropsMixin


class ToolMetadataBase(MetadataBase):
    """Factor stuff out to here."""

    @abstractmethod
    def _mk_identifier(self, **kwargs):
        pass

    @abstractmethod
    def _check_identifier(self, identifier):
        pass

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, new_identifier):
        """
        Assign an identifier to a tool. If tools_map_dict or base_dir is provided, assign an identifier that isn't in use.
        :param new_identifier:
        :return:
        """
        if self.check_index:
            if not self.tool_identifiers and self.root_repo_path:  # Populate self.tool_identifiers if possible.
                self.populate_repo_identifiers_list()
            elif not self.tool_identifiers and not self.root_repo_path:
                raise AttributeError(f"Cannot check identifiers without a root_repo_path or list of identifiers.")
            else:  # self.tool_identifiers is already set.
                pass  # debug message
        if new_identifier:
            identifier = self._check_identifier(new_identifier)  # Let it error if duplicate identifier explicitly passed.
        else:
            identifier = self._mk_identifier()
        self._identifier = identifier

    @property
    def keywords(self):
        return self._keywords

    @keywords.setter
    def keywords(self, keywords_list):
        if keywords_list:
            keywords = []
            for keyword in keywords_list:
                if isinstance(keyword, Keyword):
                    keywords.append(keyword)
                else:
                    if isinstance(keyword, dict):
                        keywords.append(Keyword(**keyword))
                    else:
                        keywords.append(Keyword(keyword))
        else:
            keywords = None
        self._keywords = keywords

    def populate_repo_identifiers_list(self):
        if not self.root_repo_path:
            raise AttributeError(f"{self} does not have a root_repo_path set.")
        identifiers_index_path = Path(self.root_repo_path) / tool_index_path
        try:
            with identifiers_index_path.open('r') as identifiers_index:
                self.tool_identifiers = identifiers_index.read().splitlines()
        except FileNotFoundError:
            raise


class ParentToolMetadata(CommonPropsMixin, SoftwarePropsMixin, ToolMetadataBase):

    @staticmethod
    def _init_metadata():
        """
        Can set default values here.
        :return:
        """
        return OrderedDict([
            ('name', None),
            ('_in_index', False),  # Specify if a supplied identifier when initializing is already expected to be in the index file. If so, it will avoid an error when using _check_identifier
            ('softwareVersion', None),
            ('root_repo_path', None),  # These need to be set before identifier.
            ('check_index', False),
            ('tool_identifiers', None),
            ('identifier', None),
            ('featureList', None),
            ('metadataStatus', 'Incomplete'),
            ('description', None),
            ('codeRepository', None),
            ('license', None),
            ('WebSite', None),
            ('contactPoint', None),
            ('publication', None),
            ('keywords', None),
            ('alternateName', None),
            ('creator', None),
            ('programmingLanguage', None),
            ('datePublished', None),
            ('downloadURL', None),
            ('extra', None),
        ])

    def _loop_mk_identifier(self, name_hash, version_hash, name_hash_start_index=0, version_hash_start_index=1):
        # Very unlikely that main part of an identifier will be repeated. Version hash is most likely culprit. Can refine later if that's a problem.
        identifier = f"{tool_identifier_prefix}_{name_hash[name_hash_start_index: name_hash_start_index + 6]}.{version_hash[version_hash_start_index:version_hash_start_index + 2]}"
        try:
            self._check_identifier(identifier)
        except AssertionError:
            identifier = self._loop_mk_identifier(name_hash, version_hash, name_hash_start_index=0, version_hash_start_index=version_hash_start_index + 1)
        return identifier

    def _check_identifier(self, identifier):
        if not parent_tool_identifier_pattern.match(identifier):
            raise ValueError(f"Tool identifier not formatted correctly: {identifier}")
        if self.check_index:
            if self._in_index:
                if not identifier in self.tool_identifiers:
                    raise NotInIndexError(f"{identifier} not found in {self.tool_identifiers}")
            else:
                if identifier in self.tool_identifiers:
                    raise InIndexError(f"")
        return identifier

    def _mk_identifier(self, name_hash_start_index=0, version_hash_start_index=0):
        if not (self.name and self.softwareVersion.versionName):
            raise ValueError(f"Name and softwareVersion must be provided to make an identifier.")
        name_hash, version_hash = _mk_hashes(self.name, self.softwareVersion.versionName)
        identifier = f"{tool_identifier_prefix}_{name_hash[name_hash_start_index:name_hash_start_index + 6]}.{version_hash[version_hash_start_index:version_hash_start_index + 2]}"
        try:
            self._check_identifier(identifier)
        except AssertionError:
            identifier = self._loop_mk_identifier(name_hash, version_hash)
        return identifier

    def make_subtool_metadata(self, subtool_name, **kwargs):
        if not self.featureList:
            raise ValueError(f"Cannot create subtool. featureList of {self.name} is not populated.")
        if subtool_name not in self.featureList:
            raise ValueError(f"{subtool_name} must be in the parent featureList")
        subtool_metadata = SubtoolMetadata(name=subtool_name, _parentMetadata=self, **kwargs)
        return subtool_metadata

    @classmethod
    def load_from_file(cls, file_path, ignore_empties=False, **kwargs):
        """

        :param file_path (Path/str): path to the file to load ParentMetadata from.
        :param ignore_empties (bool): If specified, will set empty attributes (those with subkeys but no values) to None.
        :param kwargs:
        :return:
        """
        file_path = Path(file_path)
        with file_path.open('r') as file:
            file_dict = safe_load(file)
        file_dict.update(kwargs)
        file_dict['check_index'] = kwargs.get('check_index', True)  # If not provided in kwargs. Assume metadata loaded from a file is already in the index.
        file_dict['_in_index'] = kwargs.get('_in_index', True)  # Expect is to already be indexed when loading from a file.
        try:
            file_dict['root_repo_path'] = kwargs['root_repo_path']
        except KeyError:
            if file_dict.get('check_index'):
                file_dict['root_repo_path'] = Path(*file_path.parts[:-5])  # This should be the root_repo_path relative to a parent metadata file.
        return cls(**file_dict, ignore_empties=ignore_empties)

    @classmethod
    def create_from_biotools(cls, biotools_id, version_name, subtools, **kwargs):
        biotools_kwargs = make_tool_metadata_kwargs_from_biotools(biotools_id)
        biotools_kwargs.update(kwargs)  # Overwrite any biotools kwargs with kwargs that were provided.
        biotools_kwargs['featureList'] = subtools
        biotools_kwargs['softwareVersion'] = {}
        biotools_kwargs['softwareVersion']['versionName'] = version_name
        return cls(**biotools_kwargs)

    def mk_file(self, base_dir, keys=None, replace_none=True, update_index=True):
        """

        :param base_dir: The root repo path of the repository to make the file in.
        :param keys: If specified, only the indicated keys will be written to the file.
        :param replace_none: If specified, None values will be replaced with default field such as an
        :param update_index: If specified, then the identifier will be added the index file when the file is made.
        :return:
        """
        file_path = get_tool_metadata(self.name, self.softwareVersion.versionName, subtool_name=None, parent=True, base_dir=base_dir)
        returned_path = super().mk_file(file_path, keys, replace_none)
        if update_index:
            index_file_path = self.root_repo_path / tool_index_path
            with index_file_path.open('a') as index_file:
                index_file.write(self.identifier)
                index_file.write('\n')
        return returned_path


class SubtoolMetadata(CommonPropsMixin, WorkflowLanguageStatusMixin, ToolMetadataBase):

    @staticmethod
    def _init_metadata():
        return OrderedDict([
            ('name', None),
            ('check_index', None),  # Will be set to False (default) in __init__
            ('_in_index', False),  # Only used if check_index is True. Will be set to False (default) in __init__
            ('current', False),
            ('metadataStatus', 'Incomplete'),
            ('cwlStatus', 'Incomplete'),
            ('nextflowStatus', 'Incomplete'),
            ('snakemakeStatus', 'Incomplete'),
            ('wdlStatus', 'Incomplete'),
            ('root_repo_path', None),  # These need to be set before identifier.
            ('tool_identifiers', None),
            ('identifier', None),
            ('description', None),
            ('keywords', None),
            ('alternateName', None),
            ('extra', dict(
                [
                    ('sha1', None),
                    ('dockerImage', None),
                ])),# the sha1 hash of the cwlfile associated with this subtool # the docker image pulled or referenced from this subtool's cwlfile
            ('parentMetadata', '../common/common-metadata.yaml'),  # relative path to parentMetadata
            ('_parentMetadata', None),  # ParentMetadata instance. Can be loaded from parentMetadata or set directly.
            ('_primary_file_attrs', None),  # Keep track of attributes that are set directly from kwargs and not inherited from parent.
        ])

    def __init__(self, _metadata_file_path=None, **kwargs):
        """
        Initialize SubtoolMetadata.
        :param _metadata_file_path(Path):  Path of yaml file that SubtoolMetadata is loaded from. Should not be used directly. Only used if class is initiated using 'load_from_file' method.
        :param kwargs(dict): Key:value pairs that describe subtool metadata.
        """
        ignore_empties = kwargs.pop('ignore_empties', None)
        self._parentMetadata = kwargs.get('_parentMetadata')
        check_index = kwargs.get('check_index', False)  # default value needs to be set to same as in _init_metadata.
        in_index = kwargs.get('_in_index', False)
        if self._parentMetadata:  # Will be the case if Subtool is intialized from ParentMetadata.make_subtool_metadata().
            assert isinstance(self._parentMetadata, ParentToolMetadata)
        else:
            self.parentMetadata = kwargs['parentMetadata']  # must have a path if it isn't set directly.
            self._load_parent_metadata(_metadata_file_path, root_repo_path=kwargs.get('root_repo_path'), ignore_empties=ignore_empties, check_index_parent=kwargs.get('check_index'),
                                       parent_in_index=kwargs.get('_in_index'))  # sets self._parentMetadata
        self._primary_file_attrs = []
        for k, value in kwargs.items():  # populate _primary_file_attrs
            if value:
                self._primary_file_attrs.append(k)  # keep track of kwargs supplied. ignore_empties and base_dir were popped off.
        # self._load_attrs_from_parent()
        super().__init__(**kwargs, ignore_empties=ignore_empties)

    @property
    def name(self):
        """Name of the subtool."""
        return self._name

    @name.setter
    def name(self, value):
        if not value in self._parentMetadata.featureList:
            raise ValueError(f"{value} is not in {self._parentMetadata.name} metadata featureList")
        self._name = value

    @classmethod
    def load_from_file(cls, file_path, ignore_empties=False, **kwargs):
        """Load subtool metadata into SubtoolMetadata from a yaml file.
        The Subtool populates it's metadata from ParentMetadata
        """
        file_path = Path(file_path)
        with file_path.open('r') as file:
            file_dict = safe_load(file)
        file_dict.update(kwargs)
        file_dict['parentMetadata'] = kwargs.get('parentMetadata', SubtoolMetadata._init_metadata()['parentMetadata'])  # This needs to be set when initializing. If not specified, set to default.
        file_dict['check_index'] = kwargs.get('check_index',
                                              True)  # Usually expect the identifier to be in the index already if loading from file.
        file_dict['_in_index'] = kwargs.get('_in_index',
                                            True)  # Usually expect the identifier to be in the index already if loading from file.
        try:
            file_dict['root_repo_path'] = kwargs['root_repo_path']
        except KeyError:
            if file_dict.get('check_index'):
                # file_dict['root_repo_path'] = get_base_dir_from_abs_path(file_path)
                file_dict['root_repo_path'] = Path(*file_path.parts[:-5])

        return cls(**file_dict, _metadata_file_path=file_path, ignore_empties=ignore_empties)

    def _load_parent_metadata(self, subtool_metadata_file_path, root_repo_path, ignore_empties=False, check_index_parent=True, parent_in_index=True):
        """
        Populate SubtoolMetadata._parentMetadata
        :param subtool_metadata_file_path(Path):
        :param ignore_empties(Bool):
        :return:
        """
        dir_name = subtool_metadata_file_path.parent
        full_path = dir_name / self.parentMetadata
        full_path = full_path.resolve()
        with full_path.open('r') as f:
            parent_metadata_dict = safe_load(f)
        parent_metadata_dict['root_repo_path'] = root_repo_path
        self._parentMetadata = ParentToolMetadata(**parent_metadata_dict, ignore_empties=ignore_empties, check_index=check_index_parent, _in_index=parent_in_index)

    def _load_attrs_from_parent(self):
        # initialize everything from parent. Will be overwritten anything supplied in kwargs. Doesn't do much anymore.
        parent_meta = self._parentMetadata
        self.parentMetadata = '../common/common-metadata.yaml'
        # self.identifier = self._mk_identifier()
        # self.keywords = parent_meta.keywords
        return

    def _loop_mk_identifier(self, parent_name_str, version_str, subtool_name_hash, hash_split_index=1):
        subtool_hash = subtool_name_hash[hash_split_index:hash_split_index + 2]  # shift hash slice if identifier already exists.
        identifier = f"{parent_name_str}_{subtool_hash}.{version_str}"
        try:
            self._check_identifier(identifier)
        except AssertionError:
            identifier = self._loop_mk_identifier(parent_name_str, version_str, subtool_name_hash, hash_split_index=hash_split_index + 1)
        return identifier

    def _mk_identifier(self):
        parent_name_str, version_str = self._parentMetadata.identifier.split('.', 1)
        subtool_name_hash = _mk_hashes(self.name)[0]
        subtool_hash = subtool_name_hash[:2]
        identifier = f"{parent_name_str}_{subtool_hash}.{version_str}"
        try:
            self._check_identifier(identifier)  # Can only hit AssertionError if tool_map_dict provided.
        except InIndexError:
            identifier = self._loop_mk_identifier(parent_name_str, version_str, subtool_name_hash)
        return identifier

    def _check_identifier(self, identifier):
        if self.check_index:
            if self.tool_identifiers:  # Check to make sure identifiers are not duplicated.
                if self._in_index:
                    try:
                        assert identifier in self.tool_identifiers
                    except AssertionError:
                        raise NotInIndexError(identifier)
                else:
                    try:
                        assert identifier not in self.tool_identifiers
                    except AssertionError:
                        raise InIndexError

        parent_identifier = self._parentMetadata.identifier
        if not identifier.startswith(parent_identifier[:9]):
            raise ValueError(f"Subtool identifier {identifier} does not properly correspond to parent identifier {parent_identifier}")
        if not identifier.endswith(parent_identifier[-3:]):  # should be '.xx'
            raise ValueError(
                f"Subtool identifier {identifier} does not properly correspond to parent identifier {parent_identifier}")
        if not subtool_identifier_pattern.match(identifier):
            raise ValueError(f"Tool identifier not formatted correctly: {identifier}")

        return identifier

    @classmethod
    def initialize_from_parent(cls, parent_metadata, subtool_name):
        subtool_dict = {}
        if not subtool_name in parent_metadata.featureList:
            raise ValueError(
                f"Cannot create subtool metadata. {subtool_name} is not in featureList of {parent_metadata.name}")
        subtool_dict['name'] = subtool_name
        return cls(**subtool_dict)

    def mk_completed_file(self):
        raise NotImplementedError

    def mk_instance(self, **kwargs):
        tool_name = f"{self._parentMetadata.name} {self.name}"
        tool_instance_dict = {'toolName': tool_name, 'toolVersion': self._parentMetadata.softwareVersion.versionName, 'toolIdentifier': self.identifier, '_subtoolMetadata': self}
        tool_instance_metadata = ToolInstanceMetadata(**tool_instance_dict, **kwargs)
        return tool_instance_metadata

    def mk_file(self, keys=None, replace_none=True, update_index=True):
        file_path = get_tool_metadata(self._parentMetadata.name, self._parentMetadata.softwareVersion.versionName, subtool_name=self.name, parent=False, base_dir=self.root_repo_path)
        if not file_path.parent.exists():
            file_path.parent.mkdir(parents=True)
        returned_path = super().mk_file(file_path, keys, replace_none)  # File is made here. Now need to add the key to the identifiers.
        if update_index:
            index_file_path = self.root_repo_path / tool_index_path
            with index_file_path.open('a') as index_file:
                index_file.write(self.identifier)
                index_file.write('\n')
        return returned_path


class ToolInstanceMetadata(MetadataBase):
    @staticmethod
    def _init_metadata():  # Todo: Look at cwlProv to see what kind of metadata they use for CommandLineTool run.
        return OrderedDict([
            ('toolName', None),
            ('toolVersion', None),
            ('name', None),
            ('metadataStatus', 'Incomplete'),
            ('jobStatus', 'Incomplete'),  # Describes status of a job file
            ('toolIdentifier', None),  # Identifier for subtool that this is an instance of.
            ('identifier', None),  # Identifier for the instance
            ('description', None),  # Description of what the instance does.
            ('command', None),  # Generated command. Decide on whether to implement here.
            ('inputObjects', None),
            ('outputObjects', None),
            ('extra', None),
            ('_tool_instance_file_path', None),  # Populated if class is initialized from a file.
            ('_subtoolMetadata', None)  # Store the SubtoolMetadata instance that this is an instance of.
        ])

    def __init__(self, _tool_instance_metadata_file_path=None, **kwargs):
        """
        Initialize ToolInstanceMetadata
        :param _metadata_file_path(Path): Absolute path of yaml file that SubtoolMetadata is loaded from. Should not be used directly. Only used if class is initiated using 'load_from_file' method.
        :param kwargs: Key:value pairs that describe tool instance metadata.
        """
        self._subtoolMetadata = kwargs.get('_subtoolMetadata')
        if self._subtoolMetadata:
            assert isinstance(self._subtoolMetadata, SubtoolMetadata)
        else:
            self._load_subtool_metadata(_tool_instance_metadata_file_path)
        super().__init__(**kwargs)

    def _load_subtool_metadata(self, tool_instance_metadata_path, ignore_empties=False):
        try:
            base_dir = get_base_dir_from_abs_path(tool_instance_metadata_path)
        except TypeError:
            raise
        subtool_metadata_path = get_subtool_metadata_path_from_tool_instance_metadata_path(tool_instance_metadata_path, base_dir=base_dir)

        self._subtoolMetadata = SubtoolMetadata.load_from_file(subtool_metadata_path, root_repo_path=base_dir)

        # with subtool_metadata_path.open('r') as f:
        #     subtool_metadata_dict = safe_load(f)
        # self._subtoolMetadata = SubtoolMetadata(**subtool_metadata_dict)

    @classmethod
    def load_from_file(cls, file_path, ignore_empties=False):
        file_path = Path(file_path)
        with file_path.open('r') as f:
            file_dict = safe_load(f)
        return cls(**file_dict, _tool_instance_metadata_file_path=file_path, ignore_empties=ignore_empties)

    def _mk_identifier(self):
        instance_hash = uuid.uuid4().hex[:4]
        instance_identifier = f"{self.toolIdentifier}.{instance_hash}"
        self._check_identifier(instance_identifier)  # Makes sure the toolIdentifier was correct too.
        return instance_identifier

    def _check_identifier(self, identifier):
        if not identifier.startswith(self.toolIdentifier):
            raise ValueError(f"Tool instance identifier {identifier} does not properly correspond to tool identifer {self.toolIdentifier}")
        if not tool_instance_identifier_pattern.match(identifier):
            raise ValueError(f"Tool instance identifier not formatted correctly: {identifier}")

        return identifier

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, new_identifier=None):
        if new_identifier:
            identifier = self._check_identifier(new_identifier)
        else:
            identifier = self._mk_identifier()
        self._identifier = identifier

    @property
    def inputObjects(self):
        return self._inputObjects

    @inputObjects.setter
    def inputObjects(self, input_objects_list):
        if input_objects_list:
            input_objects = []
            for input_object in input_objects_list:
                if isinstance(input_object, (IOObjectItem, IOArrayItem)):
                    pass  # ready to append
                elif isinstance(input_object, dict):
                    inp_object_keys = input_object.keys()
                    if 'identifier' in inp_object_keys:
                        input_object = IOObjectItem(**input_object)
                    elif 'objects' in inp_object_keys:
                        input_object = IOArrayItem(**input_object)
                    else:
                        raise ValueError(f"")
                else:
                    raise ValueError(f"{input_object} is not a valid value for an input object.")
                input_objects.append(input_object)
        else:
            input_objects = None
        self._inputObjects = input_objects

    @property
    def outputObjects(self):
        return self._outputObjects

    @outputObjects.setter
    def outputObjects(self, output_objects_list):
        if output_objects_list:
            output_objects = []
            for outputput_object in output_objects_list:
                if isinstance(outputput_object, (IOObjectItem, IOArrayItem)):
                    pass  # ready to append
                elif isinstance(outputput_object, dict):
                    output_object_keys = outputput_object.keys()
                    if 'identifier' in output_object_keys:
                        outputput_object = IOObjectItem(**outputput_object)
                    elif 'objects' in output_object_keys:
                        outputput_object = IOArrayItem(**outputput_object)
                    else:
                        raise ValueError(f"")
                else:
                    raise ValueError(f"{outputput_object} is not a valid value for an output object.")
                output_objects.append(outputput_object)
        else:
            output_objects = None
        self._outputObjects = output_objects

    def mk_file(self, base_dir, keys=None, replace_none=True):
        input_hash = self.identifier[-4:]
        file_path = get_tool_instance_metadata_path(self._subtoolMetadata._parentMetadata.name, self.toolVersion, input_hash=input_hash, subtool_name=self._subtoolMetadata.name, base_dir=base_dir)
        if not file_path.parent.exists():
            file_path.parent.mkdir(parents=True)
        return super().mk_file(file_path, keys, replace_none)




from unittest import skip
from tempfile import TemporaryDirectory
from tests.test_base import TestBase, test_constants
from capanno_utils.add_content import main as add_content
from capanno_utils.classes.metadata.tool_metadata import ParentToolMetadata, SubtoolMetadata, ToolInstanceMetadata
from capanno_utils.helpers.get_paths import get_tool_metadata, get_tool_instance_metadata_path


class TestMakeToolInstanceMetadata(TestBase):

    def test_make_file(self):
        parent_tool_name = 'parentName'
        subtool_name = 'subtoolName'
        tool_version = 'test_version'
        with TemporaryDirectory(prefix='tool_instance_metadata') as tmpdir:
            self.make_empty_tools_dir(tmpdir)
            add_content(['-p', tmpdir, 'tool', parent_tool_name, tool_version, subtool_name])
            subtool_path = get_tool_metadata(parent_tool_name, tool_version, subtool_name, base_dir=tmpdir)
            subtool_metadata = SubtoolMetadata.load_from_file(subtool_path)
            subtool_instance = subtool_metadata.mk_instance()
            subtool_instance.mk_file(base_dir=tmpdir)
        return

    def test_load_from_file(self):
        instance_file_args = {'tool_name': 'gawk', 'tool_version': '4.1.x', 'input_hash': '1c10', 'base_dir': self.test_content_dir}
        instance_metadata_path = get_tool_instance_metadata_path(**instance_file_args)
        tool_instance = ToolInstanceMetadata.load_from_file(instance_metadata_path)
        return

    def test_load_and_from_file_and_dump(self):
        instance_file_args = {'tool_name': 'gawk', 'tool_version': '4.1.x', 'input_hash': '1c10', 'base_dir': self.test_content_dir}
        instance_metadata_path = get_tool_instance_metadata_path(**instance_file_args)
        tool_instance = ToolInstanceMetadata.load_from_file(instance_metadata_path)
        with TemporaryDirectory(prefix='loadAndDump_') as tmpdir:
            tool_instance.mk_file(base_dir=tmpdir)
            assert True
        return



import re
import logging
from pathlib import Path
from schema_salad.validate import ValidationException
from capanno_utils.helpers.get_paths import get_tool_sources, get_tool_instance_path, get_tool_dir, get_tool_instances_dir_from_cwl_path, get_tool_cwl_from_instance_path
from capanno_utils.classes.schema_salad.schema_salad import InputsSchema

def validate_inputs_for_instance(instance_path, tool_inputs_info=None):
    """
    Need to either provide inputs_schema or cwl_tool_path
    :param instance_path:
    :param inputs_schema:
    :param cwl_tool_path:
    :return:
    """
    if not tool_inputs_info:
        tool_inputs_info = get_tool_cwl_from_instance_path(instance_path)
    if isinstance(tool_inputs_info, InputsSchema):
        inputs_schema = tool_inputs_info
    else:  # Should be path to cwl tool file, or CommandLineTool object.
        inputs_schema = InputsSchema(tool_inputs_info)

    inputs_schema.validate_inputs(instance_path)  # Will raise error if not valid.
    return

def validate_all_inputs_for_tool(cwl_tool_document_path):
    instance_file_pattern = re.compile(r'[0-9a-f]{4}\.ya?ml')

    cwl_tool_document_path = Path(cwl_tool_document_path)
    inputs_schema = InputsSchema(cwl_tool_document_path)
    instances_path = get_tool_instances_dir_from_cwl_path(cwl_tool_document_path)
    for instance_file in instances_path.iterdir():
        if  instance_file_pattern.match(instance_file.name):
            try:
                validate_inputs_for_instance(instance_file, inputs_schema)
            except ValidationException:
                print(f"{instance_file} failed validation")
                raise
    return



"""
Functions to add, delete, or modify fields in yaml files.
"""

from pathlib import Path
from ruamel.yaml import round_trip_load, YAML
from capanno_utils.helpers.get_paths import get_types_from_path
from capanno_utils.validate import validate_subtool_metadata
from capanno_utils.content_maps import make_tools_map_dict


def add_field_to_to_yaml_file(dict_to_add, yaml_file, after_key=None):
    """
    Add a new field, specified by 'dict_to_add' to the file specified by 'yaml_file' at a position after the key 'after_key'.
    """
    yaml_file = Path(yaml_file)
    assert yaml_file.is_absolute()
    yaml = YAML()
    with yaml_file.open('r') as f:
        file_contents = yaml.load(f)
    if after_key:
        key_position = list(file_contents.keys()).index(after_key)
        i = 1
        for k, v in dict_to_add.items():
            file_contents.insert(key_position + i, k, v)
            i += 1
    else:
        file_contents.update(dict_to_add)
    with yaml_file.open('w') as f:
        yaml.dump(file_contents, f)
    return


def update_subtool_metadata_files(directory_or_file, dict_to_add, after_key=None, base_dir=None):
    """
    Use with caution, this can change the contents of all metadata files in a content repository.
    """
    base_type, specific_type = get_types_from_path(directory_or_file)

    if base_type == 'tool':
        if specific_type == 'metadata':  # update a single metadata file.
            add_field_to_to_yaml_file(dict_to_add, directory_or_file, after_key=after_key)
            validate_subtool_metadata(directory_or_file)
        elif specific_type == 'base_dir':  # update all subtool metadata with new fields.
            tool_map = make_tools_map_dict(base_dir)
            for identifier, values in tool_map.items():
                if values['type'] == 'subtool':
                    metadata_path = values['metadataPath']
                    # Wrap next two lines in try/except?
                    add_field_to_to_yaml_file(dict_to_add, metadata_path,after_key=after_key)
                    validate_subtool_metadata(metadata_path)
                else: # parent metadata
                    pass
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError(f"Have not implemented updating metadata files for {base_type}")
    return



from distutils.dir_util import copy_tree  # b/c shutil.copytree does not support existing dest directories until Python 3.8.
from tempfile import NamedTemporaryFile, TemporaryDirectory
from ruamel.yaml import round_trip_dump, round_trip_load, YAML
from pathlib import Path
from unittest import skip
from tests.test_base import TestBase, test_constants
from capanno_utils.repo_config import content_repo_name
from capanno_utils.repo_config import tools_dir_name
from capanno_utils.helpers.modify_yaml_files import add_field_to_to_yaml_file, update_subtool_metadata_files

class TestModifyYamlFiles(TestBase):
    def test_add_field_to_yaml_file(self):
        test_dict = {'key0': 'value_0', 'key1': 'value_1'}
        yaml = YAML()
        update_dict = {'key_new0': 'new_value0', 'key_new1': 'new_value1'}
        with NamedTemporaryFile(mode='w', prefix='test_addfield1', suffix='.yaml') as tf1: # Create a temporary yaml file
            yaml.dump(test_dict, tf1)
            add_field_to_to_yaml_file(update_dict, tf1.name, after_key='key0')
            assert True
        return

    @skip('')
    def test_modify_yaml_file(self):
        update_dict = {'snakemakeStatus': 'Incomplete', 'nextflowStatus': 'Incomplete', 'wdlStatus': 'Incomplete'}


        with TemporaryDirectory(prefix='test_modify_yaml') as td:
            copy_tree(str(self.test_content_dir), td)
            test_file_path = Path(td) / tools_dir_name / 'gawk' / '4.1.x' / 'gawk' / 'gawk-metadata.yaml'
            update_subtool_metadata_files(test_file_path, update_dict, after_key='cwlStatus')
            assert True
        return

    # @skip('')
    def test_modify_tool_metadata(self):
        update_dict = {'snakemakeStatus': 'Incomplete', 'nextflowStatus': 'Incomplete', 'wdlStatus': 'Incomplete'}

        with TemporaryDirectory(prefix='test_modify_yaml') as td:
            temp_content_dir = f"{td}/{content_repo_name}"
            copy_tree(str(self.test_content_dir), temp_content_dir)
            test_file_path = Path(temp_content_dir) / tools_dir_name
            update_subtool_metadata_files(test_file_path, update_dict, after_key='cwlStatus', base_dir=temp_content_dir)
            assert True
        return


import argparse
import sys
from pathlib import Path
from capanno_utils.status import update_all_tools_to_released
from capanno_utils.content_maps import *


def get_parser():
    parser = argparse.ArgumentParser(description=f"Change the release status of content in {content_repo_name}")
    parser.add_argument('-p', '--root-repo-path', dest='root_path', type=Path, default=Path.cwd(),
                        help="Specify the root path of your content repo if it is not the current working directory.")
    # parser.add_argument('-q', '--quiet', dest='quiet', action='store_true', help="Silence messages to stdout")
    subparsers = parser.add_subparsers(description='', dest='source_type')

    release_tools = subparsers.add_parser('tools', help='Change the release status of tools')
    release_tools.add_argument('--release-all', action='store_true', help='Change the status of all unreleased tool to Released.')
    release_tools.add_argument('--release', action='store_true', help='NotImplemented')


    return parser


def main(argsl=None):
    if argsl is None:
        argsl = sys.argv[1:]

    parser = get_parser()
    args = parser.parse_args(argsl)

    if args.source_type == 'tools':
        if args.release_all:
            update_all_tools_to_released(base_dir=args.root_path)
        elif args.release:
            raise NotImplementedError
    else:
        raise NotImplementedError

    return

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python3

# Module to call from command line.

import sys
import argparse
from pathlib import Path
from capanno_utils.add.add_tools import add_tool, add_subtool
from capanno_utils.add.add_scripts import add_script, add_common_script_metadata
from capanno_utils.add.add_workflows import add_workflow

def get_parser():
    parser = argparse.ArgumentParser(description='Import tools and workflows from other repositories')
    parser.add_argument('-p','--root_repo_path', dest='root_path', type=Path, default=Path.cwd(), help="Specify the root path of your cwl content repo if it is not the current working directory.")
    subparsers = parser.add_subparsers(description='Specify the command to run.', dest='command')

    # add_tool parser
    addtool = subparsers.add_parser('tool', help='add a new standalone tool.')
    addtool.add_argument('tool_name', type=str, help="The name of the tool to add.")
    addtool.add_argument('version_name', type=str, help="The version of the tool to add.")
    addtool.add_argument('subtool_names', nargs='*', type=str, help="The names subtool names to initialize directories for.")
    addtool.add_argument('--biotoolsID', type=str, help='biotools id from https://bio.tools')
    addtool.add_argument('--has_primary', action='store_true', help='Tool is called directly without a subtool.')
    addtool.add_argument('--init_cwl', action='store_true', help="If specified, CWL CommandLineTool files will be intiated for the subtools and primary tool if it exists.")


    # add_subtool parser
    addsubtool = subparsers.add_parser('subtool', help='add a subtool. A parent must exist.')
    addsubtool.add_argument('tool_name', help='The primary name of the tool.')
    addsubtool.add_argument('version_name', help='The version of the tool.')
    addsubtool.add_argument('subtool_name', help="The name of the subtool. The subtool name must be present in the parent's featureList field.")
    addsubtool.add_argument('-u', '--update_featureList', action='store_true', help='Update the featureList of the Application Suite metadata to contain new subtool.')
    addsubtool.add_argument('--init_cwl', nargs='?', type=str, default=False, const=True,
                         help="If specified, CWL CommandLineTool files will be intiated for the subtools and primary tool if it exists.")

    # add_common_script_parser
    addscriptcommon = subparsers.add_parser('common_script', help='add script metadata that other scripts can inherit from')
    addscriptcommon.add_argument('group_name', help='The name of the group directory that the script will go into.')
    addscriptcommon.add_argument('project_name', help='The name of the project directory that the script will go into.')
    addscriptcommon.add_argument('script_version', help='The version of the script.')
    addscriptcommon.add_argument('filename', help="The name of common metadata file. File extensions and '-metadata.yaml' postfix must be omitted.")

    # add_script parser
    addscript = subparsers.add_parser('script', help='add a script')
    addscript.add_argument('group_name', help='The name of the group directory that the script will go into.')
    addscript.add_argument('project_name', help='The name of the project directory that the script will go into.')
    addscript.add_argument('script_version', help='The version of the script.')
    addscript.add_argument('script_name', help='The name of the script. File extensions should be omitted. Replace spaces with underscores')
    addscript.add_argument('--parent_metadata', '-p', nargs='+', help="path(s) to common script metadata that the script should inherit from.")
    addscript.add_argument('--init_cwl', action='store_true',
                         help="If specified, CWL CommandLineTool files will be intiated for the subtools and primary tool if it exists. ")
    # Could add flags for more kwargs here. Probably easiest to just add to file though.

    addworkflow = subparsers.add_parser('workflow', help='add a workflow')
    addworkflow.add_argument('group_name', help='The name of the group directory that the workflow will go into.')
    addworkflow.add_argument('workflow_name', help='The name of the workflow. Replace spaces with underscores.')
    addworkflow.add_argument('workflow_version', help='The version of the workflow.')

    return parser


def main(argsl=None):
    if not argsl:
        argsl = sys.argv[1:]
    parser = get_parser()
    args = parser.parse_args(argsl)

    if args.command == 'tool':
        add_tool(args.tool_name, args.version_name, subtool_names=args.subtool_names, biotools_id=args.biotoolsID, has_primary=args.has_primary, init_cwl=args.init_cwl, root_repo_path=args.root_path)
    elif args.command == 'subtool':
        add_subtool(args.tool_name, args.version_name, args.subtool_name, update_featureList=args.update_featureList, init_cwl=args.init_cwl, root_repo_path=args.root_path)
    elif args.command == 'common_script':
        add_common_script_metadata(args.group_name, args.project_name, args.script_version, args.filename, root_repo_path=args.root_path)
    elif args.command == 'script':
        add_script(args.group_name, args.project_name, args.script_version, args.script_name, parentMetadata=args.parent_metadata, init_cwl=args.init_cwl, root_repo_path=args.root_path)
    elif args.command == 'workflow':
        add_workflow(args.group_name, args.workflow_name, args.workflow_version)
    else:
        parser.print_help()
    return

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))


from argparse import Action
from capanno_utils.repo_config import main_tool_subtool_name


def parse_kv_list(list, delimiter='='):
    kv_dict = {}
    for item in list:
        key, value = item.split(delimiter)
        kv_dict[key] = value

    return kv_dict

class keyvalueparse(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        if values == []: # option was specified without arguments.
            setattr(namespace, self.dest, True)
        elif values == None: # option was not specified.
            setattr(namespace, self.dest, False)
        else:
            assert isinstance(values, list)
            if namespace.has_primary and len(namespace.subtool_names) == 0: # There is only a main tool.
                # import pdb; pdb.set_trace()
                if len(values) == 1:
                    values = [f"{main_tool_subtool_name}={values[0]}"]

            parsed_values = parse_kv_list(values)
            setattr(namespace, self.dest, parsed_values)



#
# * This file is subject to the terms and conditions defined in
# * file 'LICENSE.txt', which is part of this source code package.

from .shared_properties import Publication, Person, CodeRepository, WebSite, Keyword, SoftwareVersion, object_attributes
from hashlib import md5
import uuid




def _mk_hashes(arg1, *args):
    arg1 = str(arg1)
    hashes = [md5(arg1.encode(encoding='utf-8')).hexdigest()]
    for arg in args:
        arg = str(arg)
        hashes.append(md5(arg.encode(encoding='utf-8')).hexdigest())
    return hashes


def mk_empty_prop_object(property_name):
    """
    Should only call this to override a None value for a property.
    :param property_name:
    :return:
    """
    prop_map = {'codeRepository': {'name': None}, 'publication': [{'identifier': None}],
                'contactPoint': [{'name': None}], 'creator': [{'name': None}], 'WebSite': [{'name': None}], 'keywords': [Keyword()], 'applicationSuite': {'name': None, 'softwareVersion': None, 'identifier': None}}
    if not property_name in prop_map:
        return None

    return prop_map[property_name]


def is_attr_empty(attribute):
    # Need to check for lists.
    if isinstance(attribute, list):
        is_empty = True
        for item in attribute:
            if not is_attr_empty(item):
                is_empty = False
                break
    elif isinstance(attribute, object_attributes):
        is_empty = attribute.is_empty()
    elif isinstance(attribute, bool):  # True or False has been provided.
        is_empty = False
    else:
        if attribute:
            is_empty = False
        else:
            is_empty = True
    return is_empty


class WorkflowLanguageStatusMixin:

    @property
    def cwlStatus(self):
        return self._cwlStatus

    @cwlStatus.setter
    def cwlStatus(self, cwl_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not cwl_status:
            raise ValueError("cwlStatus must be set.")
        elif cwl_status not in allowed_statuses:
            raise ValueError(f"cwlStatus must be on of  {allowed_statuses}, not {cwl_status}")
        else:
            self._cwlStatus = cwl_status

    @property
    def snakemakeStatus(self):
        return self._snakemakeStatus

    @snakemakeStatus.setter
    def snakemakeStatus(self, snakemake_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not snakemake_status:
            raise ValueError("snakemakeStatus must be set.")
        elif snakemake_status not in allowed_statuses:
            raise ValueError(f"snakemakeStatus must be on of  {allowed_statuses}, not {snakemake_status}")
        else:
            self._snakemakeStatus = snakemake_status

    @property
    def wdlStatus(self):
        return self._wdlStatus

    @wdlStatus.setter
    def wdlStatus(self, wdl_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not wdl_status:
            raise ValueError("wdlStatus must be set.")
        elif wdl_status not in allowed_statuses:
            raise ValueError(f"wdlStatus must be on of  {allowed_statuses}, not {wdl_status}")
        else:
            self._wdlStatus = wdl_status

    @property
    def nextflowStatus(self):
        return self._nextflowStatus

    @nextflowStatus.setter
    def nextflowStatus(self, nf_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not nf_status:
            raise ValueError("nextflowStatus must be set.")
        elif nf_status not in allowed_statuses:
            raise ValueError(f"nextflowStatus must be on of  {allowed_statuses}, not {nf_status}")
        else:
            self._nextflowStatus = nf_status

class SoftwarePropsMixin:
    """
    This is used by ParentToolMetadata, ScriptMetadata, and WorkflowMetadata
    """
    @property
    def softwareVersion(self):
        return (self._softwareVersion)

    @softwareVersion.setter
    def softwareVersion(self, software_version_info):

        if not software_version_info:  # softwareVersion not defined. Check in parentMetadata
            raise ValueError(f"'softwareVersion must be set.")
        if isinstance(software_version_info, SoftwareVersion):
            software_version = software_version_info
        elif isinstance(software_version_info, dict):
            software_version = SoftwareVersion(**software_version_info)
            # software_version = SoftwareVersion() # Don't initialize directly with **software_version_info. Skips setters.
            # software_version.versionName, software_version.includedVersions = software_version_info['versionName'], software_version_info['includedVersions']
        else:
            raise TypeError(f"Cannot create softwareVersion with {software_version_info}")
        self._softwareVersion = software_version
        return

    @property
    def current(self):
        return self._current

    @current.setter
    def current(self, is_current):
        if not isinstance(is_current, bool):
            raise ValueError(f"current must be a boolean value, you specified {is_current}")
        self._current = is_current
        return


    @property
    def publication(self):
        return self._publication

    @publication.setter
    def publication(self, publication_list):
        if publication_list:
            publications = [Publication(**pub) for pub in publication_list]
        else:
            publications = None
        self._publication = publications

    @property
    def contactPoint(self):
        return self._contactPoint

    @contactPoint.setter
    def contactPoint(self, person_list):
        if person_list:
            if isinstance(person_list[0], dict):
                people = [Person(**person) for person in person_list]
            elif isinstance(person_list[0], Person):
                people = person_list
        else:
            people = None
        self._contactPoint = people

    @property
    def creator(self):
        return self._creator

    @creator.setter
    def creator(self, person_list):
        if person_list:
            if isinstance(person_list[0], dict):
                people = [Person(**person) for person in person_list]
            elif isinstance(person_list[0], Person):
                people = person_list
            else:
                raise NotImplementedError
        else:
            people = None
        self._creator = people

    @property
    def codeRepository(self):
        return self._codeRepository

    @codeRepository.setter
    def codeRepository(self, code_repo_info):
        if code_repo_info:
            if isinstance(code_repo_info, CodeRepository):
                code_repo = code_repo_info
            elif isinstance(code_repo_info, dict):
                code_repo = CodeRepository(**code_repo_info)
            else:
                raise TypeError(f"Cannot create codeRepository with {code_repo_info}")
        else:
            code_repo = None
        self._codeRepository = code_repo

    @property
    def WebSite(self):
        return self._website

    @WebSite.setter
    def WebSite(self, website_list):
        if website_list:
            if isinstance(website_list[0], WebSite):
                websites = website_list
            elif isinstance(website_list[0], dict):
                websites = [WebSite(**website_dict) for website_dict in website_list]
            else:
                raise TypeError(f"Cannot create WebSite with {website_list}")
        else:
            websites = None
        self._website = websites

    @property
    def license(self):
        return self._license

    @license.setter
    def license(self, license_identifier):
        self._license = license_identifier  # use spdx identifiers

class CommonPropsMixin:
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        if not new_name:
            raise ValueError(f"'name'' must be set.")
        self._name = new_name

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description_):
        # Accept markdown
        self._description = description_

    @property
    def keywords(self):
        return self._keywords

    @keywords.setter
    def keywords(self, keywords_list):
        if keywords_list:
            keywords = []
            for keyword in keywords_list:
                if isinstance(keyword, Keyword):
                    new_keyword = keyword
                else:
                    if isinstance(keyword, dict):
                        new_keyword  =  Keyword(**keyword)
                    else:
                        new_keyword = Keyword(keyword)  # Should be a uri.
                keywords.append(new_keyword)
        else:
            keywords = None
        self._keywords = keywords

    @property
    def metadataStatus(self):
        return self._metadataStatus

    @metadataStatus.setter
    def metadataStatus(self, metadata_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not metadata_status:
            raise ValueError("metadataStatus must be set.")
        elif metadata_status not in allowed_statuses:
            raise ValueError(f"metadataStatus must be on of  {allowed_statuses}, not {metadata_status}")
        else:
            self._metadataStatus = metadata_status





import re
from urllib.parse import urlparse
from capanno_utils.repo_config import *

def get_shortened_id(input_uri):
    # return the part after the '#'
    url_parse = urlparse(input_uri)
    if url_parse.fragment:
        short_id =  url_parse.fragment
    else:
        short_id = url_parse.path.split("#")[-1]
    return short_id

def get_subtool_name_from_dir_name(tool_base_name, full_tool_name):
    pattern = re.compile(f"{tool_base_name}_")
    parts = pattern.split(full_tool_name)
    assert len(parts) == 2
    subtool_name = parts[1]
    return subtool_name

def get_type_from_identifier(identifier):
    """
    Can make this more sophisticated to handle parent tools/subtools, instances, etc. Maybe return classes rather than strings?
    :param identifier:
    :return:
    """
    if identifier.startswith(worklfow_identifier_prefix):
        identifier_type = 'workflow'
    elif identifier.startswith(tool_identifier_prefix):
        identifier_type = 'tool'
    elif identifier.startswith(script_identifier_prefix):
        identifier_type = 'script'
    else:
        raise ValueError(f"Identifier {identifier} is not recognized.")
    return identifier_type





from tests.test_base import TestBase
from pathlib import Path
from capanno_utils.helpers.get_paths import get_tool_metadata
from capanno_utils.validate import metadata_validator_factory
from capanno_utils.classes.metadata.tool_metadata import ParentToolMetadata, SubtoolMetadata


class TestValidateMetadata(TestBase):

    def test_validate_tool_metadata(self):
        # validate_tool_metadata = metadata_validator_factory(ToolMetadata)
        # metadata_path = get_tool_metadata('sort', '8.25')
        # validate_tool_metadata(metadata_path)
        # Todo update.
        return

    def test_validate_tool_metadata_fail(self):
        # validate_tool_metadata = metadata_validator_factory(ToolMetadata)
        # metadata_path = Path().cwd() / 'tests/test_files/bad_gawk-metadata.yaml'
        # with self.assertRaises(ValueError):
        #     validate_tool_metadata(metadata_path)
        # Todo update.
        return

    def test_validate_parent_tool_metadata(self):
        validate_parent_metadata = metadata_validator_factory(ParentToolMetadata)
        metadata_path = get_tool_metadata('STAR', '2.5', parent=True)
        validate_parent_metadata(metadata_path)

    def test_validate_parent_tool_metadata_fail(self):
        validate_parent_metadata = metadata_validator_factory(ParentToolMetadata)
        metadata_path = get_tool_metadata('cat', '8.x', parent=True, base_dir=self.invalid_content_dir)
        with self.assertRaises(AttributeError):
            validate_parent_metadata(metadata_path)
        return

    def test_validate_subtool_metadata(self):
        validate_subtool_metadata = metadata_validator_factory(SubtoolMetadata)
        metadata_path = get_tool_metadata('STAR', '2.5', subtool_name='alignReads')
        validate_subtool_metadata(metadata_path)
        return

    def test_validate_subtool_metadata_fail(self):
        validate_subtool_metadata = metadata_validator_factory(SubtoolMetadata)
        # metadata_path = Path.cwd() / 'tests/test_files/test_content/invalid_content/cwl-tools/samtools/1.9/samtools_view/samtools-view-metadata.yaml' # Works for now.  # subtool name not in parent featurelist.
        metadata_path = get_tool_metadata('samtools', '1.x', subtool_name='view', base_dir=self.invalid_content_dir)
        assert metadata_path.exists()
        with self.assertRaises(ValueError):
            validate_subtool_metadata(metadata_path)
        return




import tempfile
import logging
from pathlib import Path
from ruamel.yaml import safe_load
from semantic_version import Version
from capanno_utils.classes.metadata.tool_metadata import ParentToolMetadata, SubtoolMetadata
from capanno_utils.classes.metadata.script_metadata import ScriptMetadata, CommonScriptMetadata
from capanno_utils.classes.metadata.workflow_metadata import WorkflowMetadata
from .content_maps import *
from .helpers.get_paths import get_metadata_path, get_base_dir, get_tool_sources_from_metadata_path, get_workflow_sources_from_metadata_path
from .helpers.validate_cwl import validate_cwl_doc
from .helpers.validate_wdl import validate_wdl_doc
from .validate_inputs import validate_all_inputs_for_tool


def metadata_validator_factory(class_to_validate):
    def metadata_validator(metadata_path):
        try:
            metadata_instance = class_to_validate.load_from_file(metadata_path)
            # print(f"Metadata in {metadata_path} is valid {str(class_to_validate)}")
            logging.info(f"Metadata in {metadata_path} is valid {str(class_to_validate)}")
        except:
            logging.error(f"{str(class_to_validate)} in {metadata_path} failed validation.")
            raise
        return

    return metadata_validator


validate_parent_tool_metadata = metadata_validator_factory(ParentToolMetadata)
validate_subtool_metadata = metadata_validator_factory(SubtoolMetadata)
validate_script_metadata = metadata_validator_factory(ScriptMetadata)
validate_common_script_metadata = metadata_validator_factory(CommonScriptMetadata)
validate_workflow_metadata = metadata_validator_factory(WorkflowMetadata)


def validate_tool_content_from_map(tool_map_dict, base_dir=None):
    """
    tool_map(dict): Keys are identifers, values are dict with path, metadataStatus, name, versionName, and type keys.
    """
    validate_statuses = ('Draft', 'Released')
    if base_dir is None:
        base_dir = get_base_dir()
    for identifier, values in tool_map_dict.items():
        metadata_path = base_dir / values['metadataPath']
        tool_type = values['type']

        if tool_type == 'parent':  # could now also get type directly from path.
            if not 'common' in metadata_path.parts:
                raise ValueError(f"")
            validate_parent_tool_metadata(metadata_path)
        else:  # is a subtool
            validate_subtool_metadata(metadata_path)
            tool_sources = get_tool_sources_from_metadata_path(metadata_path)
            cwl_path, wdl_path, sm_path, nf_path = tuple(tool_sources.values())
            cwl_status = values['cwlStatus']
            if cwl_status in validate_statuses:
                validate_cwl_doc(cwl_path)
                validate_all_inputs_for_tool(cwl_path)
            if values['wdlStatus'] in validate_statuses:
                validate_wdl_doc(wdl_path)
            if values['nextflowStatus'] in validate_statuses:
                if not nf_path.exists():
                    raise FileNotFoundError(f"{str(nf_path)} does not exist.")
                logging.info(f"Nexflow files are not validated. {nf_path}")
            if values['snakemakeStatus'] in validate_statuses:
                if not sm_path.exists():
                    raise FileNotFoundError(f"{str(sm_path)} does not exist.")
                logging.info(f"Snakemake files are not validated {sm_path}")
    return


def validate_tools_dir(base_dir=None):
    """
    Validate all cwl files, metadata files, instances and instance metadata in a tools directory
    :return:
    """

    tool_map_dict = make_tools_map_dict(base_dir=base_dir)
    validate_tool_content_from_map(tool_map_dict, base_dir)

    return


def validate_main_tool_directory(tool_name, base_dir=None):
    """
    Validate all content in a tool directory. All versions, subtools, etc.
    """
    tool_map_dict = make_main_tool_map(tool_name, base_dir=base_dir)
    validate_tool_content_from_map(tool_map_dict, base_dir)
    return


def validate_tool_version_dir(tool_name, tool_version, base_dir=None):
    tool_version_map = make_tool_version_dir_map(tool_name, tool_version, base_dir=base_dir)
    validate_tool_content_from_map(tool_version_map, base_dir=base_dir)
    return


def validate_tool_common_dir(tool_name, tool_version, base_dir=None):
    common_tool_map = make_tool_common_dir_map(tool_name, tool_version, base_dir=base_dir)
    validate_tool_content_from_map(common_tool_map, base_dir=base_dir)
    return


def validate_subtool_dir(tool_name, version_name, subtool_name=None, base_dir=None):
    subtool_dir_map = make_subtool_map(tool_name, version_name, subtool_name, base_dir=base_dir)
    validate_tool_content_from_map(subtool_dir_map, base_dir=base_dir)
    return


# Scripts stuff

def validate_script_content_from_map(script_map_dict, base_dir=None):
    if base_dir is None:
        base_dir = get_base_dir()
    for identifier, values in script_map_dict.items():
        # validate metadata
        script_path = base_dir / values['path']
        metadata_path = get_metadata_path(script_path)
        validate_script_metadata(metadata_path)

        # validate cwl
        cwl_status = values['cwlStatus']
        if cwl_status in ('Draft', 'Released'):
            validate_cwl_doc(script_path)
            validate_all_inputs_for_tool(script_path)
    return


def validate_scripts_dir(base_dir=None):
    script_map_temp_file = tempfile.NamedTemporaryFile(prefix='scripts_map', suffix='.yaml',
                                                       delete=True)  # Change to False if file doesn't persist long enough.
    make_script_maps(script_map_temp_file.name, base_dir=base_dir)
    with script_map_temp_file as script_map:
        script_map_dict = safe_load(script_map)
    validate_script_content_from_map(script_map_dict, base_dir)
    return


def validate_group_scripts_dir(group_name, base_dir=None):
    group_script_map = make_group_script_map(group_name, base_dir=base_dir)
    validate_script_content_from_map(group_script_map, base_dir=base_dir)
    return


def validate_project_scripts_dir(group_name, project_name, base_dir=None):
    project_script_map = make_project_script_map(group_name, project_name, base_dir=base_dir)
    validate_script_content_from_map(project_script_map, base_dir=base_dir)
    return


def validate_version_script_dir(group_name, project_name, version_name, base_dir=None):
    version_script_map = make_script_version_map(group_name, project_name, version_name, base_dir=base_dir)
    validate_script_content_from_map(version_script_map, base_dir=base_dir)
    return


def validate_script_dir(group_name, project_name, version_name, script_name, base_dir=None):
    script_map = make_script_map(group_name, project_name, version_name, script_name, base_dir=base_dir)
    validate_script_content_from_map(script_map, base_dir=base_dir)
    return

# ## Workflows stuff

def validate_workflows_from_map(workflow_map_dict, base_dir=None):
    if base_dir is None:
        base_dir = get_base_dir()
    for identifier, values in workflow_map_dict.items():
        workflow_metadata = base_dir / values['metadataPath']
        validate_workflow_metadata(workflow_metadata)

        wf_status = values['workflowStatus']
        if wf_status in ('Draft', 'Released'):
            wf_language = values['workflowLanguage']
            try:
                workflow_path = workflow_metadata.parent / values['workflowPath']
            except TypeError:
                print(values)
                raise
            if not workflow_path.exists():
                raise FileNotFoundError(f"{str(workflow_path)} does not exist.")
            logging.debug(
                f"Make sure you validate {workflow_path}")  # Todo. Think I have good way to validate somewhere. Need to port here (needs to be put in a temporary directory with the tools and workflows that it calls.)
    return


def validate_workflows_dir(base_dir=None):

    workflow_map_dict = make_workflow_maps_dict(base_dir=base_dir)
    validate_workflows_from_map(workflow_map_dict, base_dir)
    return

def validate_workflow_group_dir(group_name, base_dir=None):

    workflow_map_dict = make_group_workflow_map(group_name, base_dir)
    validate_workflows_from_map(workflow_map_dict, base_dir)
    return

def validate_workflow_project_dir(group_name, project_name, base_dir=None):
    workflow_map_dict = make_project_workflow_map(group_name, project_name, base_dir)
    validate_workflows_from_map(workflow_map_dict, base_dir)
    return

def validate_workflow_version_dir(group_name, project_name, version_name, base_dir=None):
    workflow_map_dict = make_version_workflow_map(group_name, project_name, version_name, base_dir)
    validate_workflows_from_map(workflow_map_dict, base_dir)
    return

# Whole repo

def validate_repo(base_dir=None):
    validate_tools_dir(base_dir=base_dir)
    validate_scripts_dir(base_dir=base_dir)
    validate_workflows_dir(base_dir=base_dir)
    return



def get_dict_from_list(dict_list, target_key_name, value):
    """
    Find a dictionary in a list dictionaries that has a particular value for a key.
    :param dict_list:
    :param target_key_name:
    :param value:
    :param parent_key: if the key you are looking for is nested in a parent field.
    :return:
    """
    items_found = 0
    target_index = -1
    for index, dict_item in enumerate(dict_list):
        # skip dict if key doesn't exist.
        curr_dict_item = dict_item
        item_value = curr_dict_item.get(target_key_name)
        if item_value:  # key exists in dict and has a value.
            if curr_dict_item[target_key_name] == value:
                my_dict = dict_item
                target_index = index
                items_found += 1
    if items_found == 0:
        raise ValueError(f"No dictionary was found for key: {target_key_name}, value: {value}")
    elif items_found > 1:
        raise ValueError(f"Found {items_found} dictionaries with key: {target_key_name}, value: {value}. Value is not unique")
    elif items_found == 1:
        return my_dict, target_index
    else:
        raise NotImplementedError("Should never hit this.")

def no_clobber_update(dict_, update_dict):
    update_dict_keys = set(update_dict.keys())
    dict_keys = set(dict_.keys())
    duplicate_keys = list(dict_keys & update_dict_keys)
    if duplicate_keys:
        duplicate_dicts = [{duplicate_key: [dict_[duplicate_key], update_dict[duplicate_key]]} for duplicate_key in duplicate_keys]
        raise ValueError(f"Duplicate keys found. {duplicate_dicts}")
    else:
        return dict_.update(update_dict)





import argparse
import sys
from pathlib import Path
from capanno_utils.classes.metadata.workflow_metadata import WorkflowMetadata
import uuid

def get_parser():
    parser = argparse.ArgumentParser(description=f"Generate a tool, script or workflow identifier")
    parser.add_argument('type', type=str,
                        help='tool, script, workflow, workflowinstance')
    parser.add_argument('name', type=str,
                        help='Provide the name of the tool, script, or workflow')
    parser.add_argument('--identifier', type=str,
                        help='Provide the parent identifier of the workflow (for new instances)', required=False)
    parser.add_argument('version',type=str,
                        help='Provide the version of the tool, script, or workflow')
    parser.add_argument('-p', '--root-repo-path', dest='root_path', type=Path, default=Path.cwd(),
                        help="Specify the root path of your cwl content repo if it is not the current working directory.")
    parser.add_argument('-q', '--quiet', dest='quiet', action='store_true', help="Silence messages to stdout")

    return parser


def main(argsl=None):
    if argsl is None:
        argsl = sys.argv[1:]

    parser = get_parser()
    args = parser.parse_args(argsl)
    if args.type == 'workflow':
        kwargs = {"name": args.name, "softwareVersion": {"versionName": args.version, "includedVersions": args.version}, "cwlStatus": "Released"}
        wf = WorkflowMetadata(**kwargs)
        print(wf._mk_identifier())
    elif args.type == 'workflowinstance':
        if args.identifier:
            uuid_string = uuid.uuid4().hex[:4]
            print(f"{args.identifier}.{uuid_string}")
        else:
            kwargs = {"name": args.name, "softwareVersion": {"versionName": args.version, "includedVersions": args.version}, "cwlStatus": "Released"}
            wf = WorkflowMetadata(**kwargs)
            print(wf._make_workflow_inst_identifier())

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))






import os
from tests.test_base import TestBase
from capanno_utils.repo_config import config
from capanno_utils.validate import validate_tools_dir, validate_scripts_dir, validate_repo


class TestValidateDirectories(TestBase):

    def test_validate_tools_dir(self):
        validate_tools_dir(base_dir=config[os.environ['CONFIG_KEY']]['base_path'])
        return

    def test_validate_scripts_dir(self):
            validate_scripts_dir(base_dir=config[os.environ['CONFIG_KEY']]['base_path'])
            return

    # def test_validate_repo(self):
    #     validate_repo(base_dir=config[os.environ['CONFIG_KEY']]['base_path'])
    #     return



import sys
import shutil
from pathlib import Path
from ruamel.yaml import safe_load, dump
from tests.test_base import TestBase
from capanno_utils.content_maps import make_tools_index
from capanno_utils.helpers.get_paths import get_tool_version_dir, get_root_tools_dir, get_tool_common_dir, get_script_version_dir, get_root_scripts_dir


tool_args_list = [('cat', '8.x'), ('samtools', '1.x'), ('gawk', '4.1.x'), ('sort', '8.x'), ('STAR', '2.5'), ('STAR', '2.7.x'), ('md5sum', '8.x')]

script_args_list = [
    ('ENCODE-DCC', 'atac-seq-pipeline', '1.1.x'),
    ]

workflow_args = [
    ('example_workflows', 'cat_sort', 'master'),
    ]

src_content_dir = TestBase.src_content_dir
test_content_dir = TestBase.test_content_dir
invalid_content_dir = TestBase.invalid_content_dir


def teardown_tools():
    """Just delete whole tools directory. Removes anything that was manually copied too."""
    test_tools_dir = get_root_tools_dir(base_dir=test_content_dir)
    invalid_tools_dir = get_root_tools_dir(base_dir=invalid_content_dir)
    if test_tools_dir.exists():
        print(f"Removing {test_tools_dir}")
        shutil.rmtree(str(test_tools_dir))
    if invalid_tools_dir.exists():
        print(f"Removing {invalid_tools_dir}")
        shutil.rmtree(str(invalid_tools_dir))
    return

def copy_tools():
    for tool_args in tool_args_list:
        src_tool_dir = get_tool_version_dir(*tool_args, base_dir=src_content_dir)
        dest_tool_dir = get_tool_version_dir(*tool_args, base_dir=test_content_dir)
        shutil.copytree(src_tool_dir, dest_tool_dir)
    make_tools_index(base_dir=test_content_dir)
    return

def copy_to_invalid_tools():
    for tool_args in tool_args_list:
        src_tool_dir = get_tool_version_dir(*tool_args, base_dir=src_content_dir)
        invalid_tool_dir = get_tool_version_dir(*tool_args, base_dir=invalid_content_dir)
        shutil.copytree(src_tool_dir, invalid_tool_dir)
    make_tools_index(base_dir=invalid_content_dir)
    return


def update_invalid_tools():
    cat_common_dir = get_tool_common_dir(*tool_args_list[0], base_dir=invalid_content_dir)
    cat_common_metadata_path = cat_common_dir / 'common-metadata.yaml'
    with cat_common_metadata_path.open('r') as metadata_file:
        metadata_dict = safe_load(metadata_file)
    cat_common_metadata_path.unlink()  # Delete the original.
    metadata_dict['featurList'] = metadata_dict.pop('featureList')
    with cat_common_metadata_path.open('w') as metadata_file:
        dump(metadata_dict, metadata_file)


    samtools_common_dir = get_tool_common_dir(*tool_args_list[1], base_dir=invalid_content_dir)
    samtools_subtool = 'view'
    samtools_common_metadata_path = samtools_common_dir / 'common-metadata.yaml'
    with samtools_common_metadata_path.open('r') as metadata_file:
        samtools_meta_dict = safe_load(metadata_file)
    samtools_common_metadata_path.unlink()
    samtools_meta_dict['featureList'].remove(samtools_subtool)
    with samtools_common_metadata_path.open('w') as metadata_file:
        dump(samtools_meta_dict, metadata_file)
    return

def populate_invalid_tools():
    copy_to_invalid_tools()
    update_invalid_tools()
    return


def teardown_scripts():
    test_scripts_dir = get_root_scripts_dir(base_dir=test_content_dir)
    if test_scripts_dir.exists():
        shutil.rmtree(test_scripts_dir)
    return

def copy_scripts():
    for script_args in script_args_list:
        src_script_dir = get_script_version_dir(*script_args, base_dir=src_content_dir)
        dest_script_dir = get_script_version_dir(*script_args, base_dir=test_content_dir)
        shutil.copytree(src_script_dir, dest_script_dir)
    return

def populate_scripts():
    teardown_scripts()
    copy_scripts()

def make_content_maps():
    raise NotImplementedError



def main():
    teardown_tools()
    copy_tools()
    populate_invalid_tools()
    populate_scripts()

if  __name__ == "__main__":
    print(f"Updating {test_content_dir}")
    sys.exit(main())


#
# * This file is subject to the terms and conditions defined in
# * file 'LICENSE.txt', which is part of this source code package.


from abc import ABC, abstractmethod
from tempfile import NamedTemporaryFile
import logging
from pathlib import Path
import semantic_version
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml import YAML
from ...classes.metadata.common_functions import mk_empty_prop_object, is_attr_empty
from ...classes.metadata.shared_properties import object_attributes
    # CodeRepository, Person, Publication, WebSite, Keyword, ApplicationSuite, ParentScript, Tool, IOObjectItem, CallMap
#
# object_attributes = (CodeRepository, Person, Publication, WebSite, Keyword, ApplicationSuite, ParentScript, Tool, IOObjectItem, CallMap)

class MetadataBase(ABC):
    """Factor stuff out to here."""

    @staticmethod
    @abstractmethod
    def _init_metadata():
        return dict([('name', None), ('metadataStatus', None)])

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        # Can put validators here.
        self._name = new_name

    @property
    def root_repo_path(self):
        return self._root_repo_path

    @root_repo_path.setter
    def root_repo_path(self, new_root_repo_path):
        if new_root_repo_path:
            self._root_repo_path = Path(new_root_repo_path)
        else:
            self._root_repo_path = None
        return

    @property
    def repo_map_dict(self):
        return self._repo_map_dict

    @repo_map_dict.setter
    def repo_map_dict(self, new_repo_map_dict):
        self._repo_map_dict = new_repo_map_dict
        return

    @property
    def metadataStatus(self):
        return self._metadataStatus

    @metadataStatus.setter
    def metadataStatus(self, metadata_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not metadata_status:
            raise ValueError(f"metadataStatus must be set.")
        elif metadata_status not in allowed_statuses:
            raise ValueError(f"metadataStatus cannot be '{metadata_status}' must be one of {allowed_statuses}")
        else:
            self._metadataStatus = metadata_status


    def _get_metafile_keys(self):
        return list(self._init_metadata())

    def __init__(self, **kwargs):
        init_metadata = self._init_metadata()
        ignore_empties = kwargs.pop('ignore_empties', None)  # if not there will be None which is falsey so default is not to ignore.
        # kwargs.pop('ignore_empties', None)
        for k, v in kwargs.items():
            if not k in init_metadata:
                raise AttributeError(f"{k} is not a valid key for {type(self)}")

        for k, v in init_metadata.items():
            if k in kwargs:
                setattr(self, k, kwargs[k])  # Highest priority.
            else:
                try:
                    if getattr(self, k):  # value has already been set by derived class __init__. Second highest priority.
                        continue
                    elif getattr(self, k) == None:  # None values are okay.
                        continue
                    else:  # Have an empty dict, list, or something.
                        raise NotImplementedError(f"Figure out what's happening here and fix it.")
                except AttributeError:
                    setattr(self, k, v)  # Set to default value provided in self._init_metadata. Last resort.
            if ignore_empties:   # Set attribute to None rather than have empty subfields.
                attribute = getattr(self, k)
                if is_attr_empty(attribute):
                    # print(f"Setting {k} to None for {self.name}")
                    setattr(self, k, None)

        return

    def mk_file(self, file_path, keys=None, replace_none=True):
        """

        :param file_path:
        :param keys: if provided, only the keys specified will be included in the file.
        :param replace_none: Will replace None values with empty objects (keys without values) so values can be easily added in the file.
        :return:
        """
        file_path = Path(file_path)
        meta_map = CommentedMap()
        if not keys:
            keys = self._get_metafile_keys()
        for key in keys:
            if key.startswith('_'):
                continue
            if key in ('root_repo_path', 'repo_map_dict', 'check_index', 'tool_identifiers'):
                continue
            if getattr(self, key) is None:
                if replace_none:
                    setattr(self, key, mk_empty_prop_object(key))
                else:
                    continue  # Don't include keys with None values in the file.
            attr_value = getattr(self, key)
            if isinstance(attr_value, object_attributes):
                meta_map[key] = attr_value.dump()
            elif isinstance(attr_value, list):
                if not attr_value:  # empty list
                    meta_map[key] = attr_value
                elif isinstance(attr_value[0], object_attributes):  # Will not work for nested lists.
                    meta_map[key] = [item.dump() for item in attr_value]
                else:
                    meta_map[key] = attr_value
            else:
                meta_map[key] = attr_value
        yaml = YAML()
        yaml.default_flow_style = False
        yaml.indent(mapping=2, sequence=4, offset=2)

        with file_path.open('w') as yaml_file:
                yaml.dump(meta_map, yaml_file)
        assert file_path.exists()
        with file_path.open('r') as blah:
            logging.debug(f"{file_path} contents: {blah.readlines()}")
            # print(f"{file_path} contents: {blah.readlines()}")
        return file_path



#
# * This file is subject to the terms and conditions defined in
# * file 'LICENSE.txt', which is part of this source code package.

from collections import OrderedDict
import uuid
from pathlib import Path
import re
from abc import abstractmethod
from ruamel.yaml import safe_load
from capanno_utils.repo_config import *
from ...classes.metadata.shared_properties import CodeRepository, WebSite, Person, Publication, Keyword, CallMap
from ...classes.metadata.common_functions import _mk_hashes, CommonPropsMixin, SoftwarePropsMixin
from ...classes.metadata.metadata_base import MetadataBase

class WorkflowMetadataBase(MetadataBase):
    @abstractmethod
    def _mk_identifier(self, **kwargs):
        pass

    @abstractmethod
    def _check_identifier(self, identifier):
        pass

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, identifier=None, **kwargs):
        if identifier:
            identifier = self._check_identifier(identifier)
        else:
            identifier = self._mk_identifier(**kwargs)
        self._identifier = identifier

    @property
    def keywords(self):
        return self._keywords

    @keywords.setter
    def keywords(self, keywords_list):
        if keywords_list:
            keywords = []
            for keyword in keywords_list:
                if isinstance(keyword, Keyword):
                    keywords.append(keyword)
                else:
                    if isinstance(keyword, dict):
                        keywords.append(Keyword(**keyword))
                    else:
                        keywords.append(Keyword(keyword))
        else:
            keywords = [Keyword()]
        self._keywords = keywords



class WorkflowMetadata(CommonPropsMixin, SoftwarePropsMixin, WorkflowMetadataBase):

    @staticmethod
    def _init_metadata():
        return OrderedDict([
        ('name', None),
        ('softwareVersion', None),
        ('current', False),
        ('description', None),
        ('identifier', None),
        ('metadataStatus', 'Incomplete'),
        ('workflowStatus', 'Incomplete'),
        ('workflowLanguage', 'wdl'),
        ('workflowFile', None), # Workflow file.
        ('repoName', None),
        ('gitTag', None),
        ('inputsTemplate', None),
        ('callMap', None),  # This field is to make associations between called tasks/steps and underlying tools/scripts.
        ('graphStatus', 'Incomplete'),
        ('executable', False),  # if configured to execute on Truwl.
        ('codeRepository', None),
        ('WebSite', None),
        ('license', None),
        ('contactPoint', None),
        ('publication', None),
        ('keywords', None),
        ('alternateName', None),
        ('creator', None),
        ('datePublished', None),
    ])

    def _check_identifier(self, identifier):
        if not identifier[:3] == f"{worklfow_identifier_prefix}_":
            raise ValueError(f"Workflow identifiers must start with '{worklfow_identifier_prefix}_' you provided {identifier}")
        else:
            hex_pattern = r'[0-9a-f]{6}\.[0-9a-f]{2,3}$'
            match_obj = re.match(hex_pattern, identifier[3:])
            if not match_obj:
                raise ValueError(f"Workflow identifier not formatted correctly: {identifier}")
        return identifier

    def _mk_identifier(self, start=0):
        if not (self.name and self.softwareVersion):
            raise ValueError(f"Name and softwareVersion must be provided to make an identifier.")
        name_hash, version_hash = _mk_hashes(self.name, self.softwareVersion)
        identifier = f"{worklfow_identifier_prefix}_{name_hash[start:start + 6]}.{version_hash[:2]}"
        return identifier

    def _make_workflow_inst_identifier(self):
        uuid_string = uuid.uuid4().hex[:4]
        workflowid = self._mk_identifier
        return f"{workflowid}.{uuid_string}"

    @property
    def workflowStatus(self):
        return self._workflowStatus

    @workflowStatus.setter
    def workflowStatus(self, wf_status):
        allowed_statuses = ('Incomplete', 'Draft', 'Released')
        if not wf_status:
            raise ValueError("workflowStatus must be set.")
        elif wf_status not in allowed_statuses:
            raise ValueError(f"worflowStatus must be on of  {allowed_statuses}, not {wf_status}")
        else:
            self._workflowStatus = wf_status

    @property
    def workflowLanguage(self):
        return self._workflowLanguage

    @workflowLanguage.setter
    def workflowLanguage(self, wf_language):
        allowed_languages = ('cwl', 'wdl', 'nextflow', 'snakemake')
        if not wf_language:
            raise ValueError("workflowLanguage must be set.")
        elif wf_language not in allowed_languages:
            raise ValueError(f"workflowLanguage must be on of  {allowed_languages}, not {wf_language}")
        else:
            self._workflowLanguage = wf_language


    @classmethod
    def load_from_file(cls, file_path, ignore_empties=False):
        file_path = Path(file_path)
        with file_path.open('r') as file:
            file_dict = safe_load(file)
        try:
            return cls(**file_dict, ignore_empties=ignore_empties)
        except AttributeError as e:
            raise Exception(f"{e.args} in {file_path}") from e

    def make_instance(self):
        raise NotImplementedError


from unittest import skip
from tests.test_base import TestBase
from capanno_utils.repo_config import config
from capanno_utils.helpers.get_paths import *
from capanno_utils.validate_content import main as validate_content


# @skip('')
class TestValidateTools(TestBase):

    def test_validate_main_tool_dir(self):
        tool_name = 'STAR'
        test_repo_path = self.test_content_dir
        main_tool_dir = get_main_tool_dir(tool_name, base_dir=test_repo_path)
        validate_content([str(main_tool_dir), '-p', str(test_repo_path), '-q'])
        return

    def test_validate_tool_version_dir(self):
        tool_name = 'samtools'
        version_name = '1.x'
        tool_version_dir = get_tool_version_dir(tool_name, version_name, base_dir=self.test_content_dir)
        validate_content([str(tool_version_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_subtool_dir(self):
        tool_name = 'md5sum'
        version_name = '8.x'
        subtool_name = 'check'
        subtool_dir = get_tool_dir(tool_name, version_name, subtool_name, base_dir=self.test_content_dir)
        validate_content([str(subtool_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_main_subtool_dir(self):
        tool_name = 'md5sum'
        version_name = '8.x'
        subtool_name = None
        subtool_dir = get_tool_dir(tool_name, version_name, subtool_name, base_dir=self.test_content_dir)
        validate_content([str(subtool_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_common_dir(self):
        tool_name = 'md5sum'
        version_name = '8.x'
        common_dir = get_tool_common_dir(tool_name, version_name, base_dir=self.test_content_dir)
        validate_content([str(common_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_instance_dir(self):
        tool_name = 'gawk'
        version_name = '4.1.x'
        subtool_name = None
        instances_dir = get_tool_instances_dir(tool_name, version_name, subtool_name=subtool_name,
                                               base_dir=self.test_content_dir)
        validate_content([str(instances_dir), '-p', str(self.test_content_dir), '-q'])
        return

# @skip('')
class TestValidateScripts(TestBase):

    def test_validate_group_script_dir(self):
        group_name = 'ENCODE-DCC'
        script_group_dir = get_script_group_dir(group_name, base_dir=self.test_content_dir)
        validate_content([str(script_group_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_project_script_dir(self):
        group_name = 'ENCODE-DCC'
        project_name = 'atac-seq-pipeline'
        script_project_dir = get_script_project_dir(group_name, project_name, base_dir=self.test_content_dir)
        validate_content([str(script_project_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_version_script_dir(self):
        group_name = 'ENCODE-DCC'
        project_name = 'atac-seq-pipeline'
        version_name = '1.1.x'
        script_version_dir = get_script_version_dir(group_name, project_name, version_name,
                                                    base_dir=self.test_content_dir)
        validate_content([str(script_version_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_script_dir(self):
        group_name = 'ENCODE-DCC'
        project_name = 'atac-seq-pipeline'
        version_name = '1.1.x'
        script_name = 'encode_ataqc'
        script_dir = get_script_dir(group_name, project_name, version_name, script_name,
                                    base_dir=self.test_content_dir)
        validate_content([str(script_dir), '-p', str(self.test_content_dir), '-q'])
        return

    def test_validate_script_instance_dir(self):
        group_name = 'ENCODE-DCC'
        project_name = 'atac-seq-pipeline'
        version_name = '1.1.x'
        script_name = 'encode_bam2ta'  # No instances in there, but might as well have a test.
        script_instance_dir = get_script_instance_dir(group_name, project_name, version_name, script_name,
                                                      base_dir=self.test_content_dir)
        validate_content([str(script_instance_dir), '-p', str(self.test_content_dir), '-q'])
        return



#!/usr/bin/env python3

import argparse
import sys
import logging
from pathlib import Path
from capanno_utils.validate import *
from capanno_utils.validate_inputs import validate_inputs_for_instance
from capanno_utils.helpers.validate_cwl import validate_cwl_doc
from capanno_utils.helpers.validate_wdl import validate_wdl_doc
from capanno_utils.helpers.get_paths import get_types_from_path


def get_parser():
    parser = argparse.ArgumentParser(description="Validate metadata and workflow language files.")
    parser.add_argument('path', type=Path,
                        help='Provide the path to validate. If a directory is specified, all content in the directory will be validated. If a file is specified, only that file will be validated.')
    parser.add_argument('-p', '--root-repo-path', dest='root_path', type=Path, default=Path.cwd(),
                        help="Specify the root path of your content repo if it is not the current working directory.")
    parser.add_argument('-q', '--quiet', dest='quiet', action='store_true', help="Silence messages to stdout")

    return parser


def main(argsl=None):
    if not argsl:
        argsl = sys.argv[1:]

    parser = get_parser()
    args = parser.parse_args(argsl)

    # from pdb import set_trace; set_trace()
    if args.path.is_absolute():
        full_path = args.path
    else:
        full_path = args.root_path / args.path

    base_type, specific_type = get_types_from_path(full_path, root_repo_name=args.root_path.name,
                                                   base_path=args.root_path)

    if not args.quiet:
        print(f"Validating {str(full_path)} \n")

    if base_type == 'tool':
        # Check for file types.
        if specific_type == 'common_metadata':
            validate_parent_tool_metadata(full_path)
        elif specific_type == 'cwl':  # Todo add validation for other files.
            validate_cwl_doc(full_path)
        elif specific_type == 'wdl':
            validate_wdl_doc(full_path)
        elif specific_type == 'metadata':
            validate_subtool_metadata(full_path)
        elif specific_type == 'instance':
            validate_inputs_for_instance(full_path)
        elif specific_type == 'instance_metadata':
            raise NotImplementedError
        # Check for directory types.
        elif specific_type == 'base_dir':
            validate_tools_dir(base_dir=args.root_path)
        elif specific_type == 'tool_dir':
            tool_name = full_path.parts[-1]
            validate_main_tool_directory(tool_name, base_dir=args.root_path)
        elif specific_type == 'version_dir':
            tool_name, version_name = full_path.parts[-2:]
            validate_tool_version_dir(tool_name, version_name, base_dir=args.root_path)
        elif specific_type == 'common_dir':
            tool_name, version_name = full_path.parts[-3:-1]
            validate_tool_common_dir(tool_name, version_name, base_dir=args.root_path)
        elif specific_type == 'subtool_dir':
            path_parts = full_path.parts
            tool_name, version_name = path_parts[-3:-1]
            subtool_name = path_parts[-1][len(tool_name) + 1:]
            if subtool_name == '':
                subtool_name = None
            validate_subtool_dir(tool_name, version_name, subtool_name, base_dir=args.root_path)
        elif specific_type == 'instance_dir':  # Must do the same as validating a subtool directory. Could skip validating subtool metadata, but won't. Don't see use for that.
            path_parts = full_path.parts
            tool_name, version_name = path_parts[-4:-2]
            subtool_name = path_parts[-2][len(tool_name) + 1:]
            if subtool_name == '':
                subtool_name = None
            validate_subtool_dir(tool_name, version_name, subtool_name=subtool_name, base_dir=args.root_path)
        else:
            raise ValueError(f"Cannot validate tool path {full_path}")
    elif base_type == 'script':
        if specific_type == 'cwl':  # Todo. Add support for other wf types.
            validate_cwl_doc(full_path)
        elif specific_type == 'metadata':
            validate_script_metadata(full_path)
        elif specific_type == 'instance':
            validate_inputs_for_instance(full_path)
        elif specific_type == 'instance_metadata':
            raise NotImplementedError
        # Check for directory types.
        elif specific_type == 'base_dir':
            validate_scripts_dir(base_dir=args.root_path)
        elif specific_type == 'group_dir':
            group_name = full_path.parts[-1]
            validate_group_scripts_dir(group_name, base_dir=args.root_path)
        elif specific_type == 'project_dir':
            group_name, project_name = full_path.parts[-2:]
            validate_project_scripts_dir(group_name, project_name, base_dir=args.root_path)
        elif specific_type == 'version_dir':
            group_name, project_name, version_name = full_path.parts[-3:]
            validate_version_script_dir(group_name, project_name, version_name, base_dir=args.root_path)
        elif specific_type == 'script_dir':
            group_name, project_name, version_name, script_name = full_path.parts[-4:]
            validate_script_dir(group_name, project_name, version_name, script_name, base_dir=args.root_path)
        elif specific_type == 'instance_dir':
            group_name, project_name, version_name, script_name = full_path.parts[-5:-1]
            validate_script_dir(group_name, project_name, version_name, script_name, base_dir=args.root_path)
        else:
            raise ValueError(f"Cannot validate script path {full_path}")

    elif base_type == 'workflow':
        if specific_type == 'base_dir':
            validate_workflows_dir(base_dir=args.root_path)
        elif specific_type == 'group_dir':
            group_name = full_path.parts[-1]
            validate_workflow_group_dir(group_name, base_dir=args.root_path)
        elif specific_type == 'project_dir':
            group_name, project_name = full_path.parts[-2:]
            validate_workflow_project_dir(group_name, project_name, base_dir=args.root_path)
        elif specific_type == 'version_dir':
            group_name, project_name, version_name = full_path.parts[-3:]
            validate_workflow_version_dir(group_name, project_name, version_name, base_dir=args.root_path)
        elif specific_type == 'cwl':  # Todo. Add other wf language types.
            raise NotImplementedError
        elif specific_type == 'metadata':
            validate_workflow_metadata(full_path)
        elif specific_type == 'instance':
            raise NotImplementedError
        elif specific_type == 'instance_metadata':
            raise NotImplementedError
        else:
            raise ValueError(f"Cannot validate workflow path {full_path}")
    elif base_type == 'repo_root':
        validate_repo(full_path)

    else:
        parser.print_help()

    if not args.quiet:
        print(f"{full_path} is valid.")
    return


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import unittest
import logging
import os
import argparse
from unittest import defaultTestLoader, TestSuite
from tests.test_add_content import TestAddToolMain, TestAddScriptMain
from tests.test_add_tool import TestAddTool
from tests.test_add_tool_instance import TestAddToolInstance
from tests.test_add_workflows import TestAddWorkflow
from tests.test_change_status import TestChangeStatus
from tests.test_dict_tools import TestDictTools
from tests.test_dump_cwl import TestDumpCwlTool
from tests.test_input_templates import TestMakeCommandLineToolInputsTemplate
from tests.test_tool_instance_metadata import TestMakeToolInstanceMetadata
from tests.test_tool_metadata import TestMakeParentToolMetadata, TestMakeSubtoolMetadata
from tests.test_script_metadata import TestScriptMetadata
from tests.test_content_maps import TestToolMaps
from tests.test_modify_yaml_files import TestModifyYamlFiles
from tests.test_path_tools import TestGetTypesFromPath
from tests.test_workflow_metadata import TestWorkflowMetadata
from tests.test_validate import TestValidateMetadata
from tests.test_validate_all import TestValidateDirectories
from tests.test_validate_all_metadata_in_maps import TestValidateContent
import tests.test_validate_content
from tests.test_validate_tool_inputs import TestValidateInputs


def suite_full():
    suite = TestSuite()
    suite.addTest(suite_add_content())
    suite.addTest(suite_add_tool())
    suite.addTest(suite_add_workflow())
    suite.addTest(suite_add_tool_instance())
    suite.addTest(suite_content_maps())
    suite.addTest(suite_dict_tools()),
    suite.addTest(suite_dump_cwl()),
    suite.addTest(suite_input_templates())
    suite.addTest(suite_script_metadata())
    suite.addTest(suite_tool_metadata())
    suite.addTest(suite_tool_instance_metadata())
    suite.addTest(suite_validate())
    suite.addTest(suite_validate_all_metadata_in_maps())
    suite.addTest(suite_validate_content())
    suite.addTest(suite_validate_directories())
    suite.addTest(suite_validate_tool_inputs())
    suite.addTest(suite_workflow_metadata())
    return suite


def suite_add_content():
    suite = defaultTestLoader.loadTestsFromTestCase(TestAddToolMain)
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(TestAddScriptMain))
    return suite


def suite_add_tool():
    suite = defaultTestLoader.loadTestsFromTestCase(TestAddTool)
    return suite


def suite_add_tool_instance():
    suite = defaultTestLoader.loadTestsFromTestCase(TestAddToolInstance)
    return suite

def suite_add_workflow():
    suite = defaultTestLoader.loadTestsFromTestCase(TestAddWorkflow)
    return suite

def suite_change_status():
    """
    Keep this test out of suite_full!! It is very long and should be used in isolation if needed.
    """
    suite = defaultTestLoader.loadTestsFromTestCase(TestChangeStatus)
    return suite

def suite_content_maps():
    suite = defaultTestLoader.loadTestsFromTestCase(TestToolMaps)
    return suite


def suite_dict_tools():
    suite = defaultTestLoader.loadTestsFromTestCase(TestDictTools)
    return suite


def suite_dump_cwl():
    suite = defaultTestLoader.loadTestsFromTestCase(TestDumpCwlTool)
    return suite


def suite_input_templates():
    suite = defaultTestLoader.loadTestsFromTestCase(TestMakeCommandLineToolInputsTemplate)
    return suite

def suite_modify_yaml_files():
    suite = defaultTestLoader.loadTestsFromTestCase(TestModifyYamlFiles)
    return suite

def suite_path_tools():
    suite = defaultTestLoader.loadTestsFromTestCase(TestGetTypesFromPath)
    return suite


def suite_script_metadata():
    suite = defaultTestLoader.loadTestsFromTestCase(TestScriptMetadata)
    return suite


def suite_tool_instance_metadata():
    suite = defaultTestLoader.loadTestsFromTestCase(TestMakeToolInstanceMetadata)
    return suite


def suite_tool_metadata():
    suite = defaultTestLoader.loadTestsFromTestCase(TestMakeParentToolMetadata)
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(TestMakeSubtoolMetadata))
    return suite


def suite_validate():
    suite = defaultTestLoader.loadTestsFromTestCase(TestValidateMetadata)
    return suite


def suite_validate_directories():
    suite = defaultTestLoader.loadTestsFromTestCase(TestValidateDirectories)
    return suite


def suite_validate_all_metadata_in_maps():
    suite = defaultTestLoader.loadTestsFromTestCase(TestValidateContent)
    return suite


def suite_validate_content():
    suite = defaultTestLoader.loadTestsFromTestCase(tests.test_validate_content.TestValidateTools)
    suite.addTest(defaultTestLoader.loadTestsFromTestCase(tests.test_validate_content.TestValidateScripts))
    return suite


def suite_validate_tool_inputs():
    suite = defaultTestLoader.loadTestsFromTestCase(TestValidateInputs)
    return suite


def suite_workflow_metadata():
    suite = defaultTestLoader.loadTestsFromTestCase(TestWorkflowMetadata)
    return suite


def suite_dict():
    suite_dict = {'add_content': suite_add_content(),
                  'add_tool': suite_add_tool(),
                  'add_workflow': suite_add_workflow(),
                  'add_tool_instance': suite_add_tool_instance(),
                  'change_status': suite_change_status(),
                  'content_maps': suite_content_maps(),
                  'dict_tools': suite_dict_tools(),
                  'dump_cwl': suite_dump_cwl(),
                  'full': suite_full(),
                  'input_templates': suite_input_templates(),
                  'modify_yaml': suite_modify_yaml_files(),
                  'path_tools': suite_path_tools(),
                  'script_metadata': suite_script_metadata(),
                  'tool_instance_metadata': suite_tool_instance_metadata(),
                  'tool_metadata': suite_tool_metadata(),
                  'validate': suite_validate(),
                  'validate_all_metadata_in_maps': suite_validate_all_metadata_in_maps(),
                  'validate_content': suite_validate_content(),
                  'validate_directories': suite_validate_directories(),
                  'validate_tool_inputs': suite_validate_tool_inputs(),
                  'workflow_metadata': suite_workflow_metadata(),
                  }
    return suite_dict


parser = argparse.ArgumentParser(description="Run test suite")
parser.add_argument("module", nargs='?', default='full', choices=suite_dict().keys(),
                    help='Specify the test suite that you would like to run. All suites will run if not specified')
parser.add_argument("--log_level", help="Set logging level", choices=['debug', 'info', 'warning', 'error', 'critical'])
args = parser.parse_args()


def run_tests():
    # logging.basicConfig(level=logging.CRITICAL)
    if args.log_level:
        if args.log_level == 'debug':
            log_level = logging.DEBUG
        elif args.log_level == 'info':
            log_level = logging.INFO
        elif args.log_level == 'warning':
            log_level = logging.WARNING
        elif args.log_level == 'error':
            log_level = logging.ERROR
        elif args.log_level == 'critical':
            log_level = logging.CRITICAL
        else:
            raise ValueError  # Should never hit this.
    else:
        log_level = logging.CRITICAL  # default level.

    logging.getLogger().setLevel(log_level)

    # update this dictionary when new suites are added.

    try:
        test_suite = suite_dict()[args.module]
    except KeyError:
        raise ValueError(f"Test suite {args.module} not recognized.")

    # Set temp directory

    os.environ['CONFIG_KEY'] = 'TEST'
    unittest.TextTestRunner().run(test_suite)
    return


if __name__ == '__main__':
    run_tests()




from tests.test_base import TestBase
from tempfile import NamedTemporaryFile
from capanno_utils import repo_config
from capanno_utils.helpers.get_paths import get_tool_sources
from capanno_utils.classes.cwl.common_workflow_language import load_document

class TestDumpCwlTool(TestBase):

    def test_dump_to_file(self):
        tool_name = 'cat'
        tool_version = '8.x'
        subtool_name = repo_config.main_tool_subtool_name
        tool_sources = get_tool_sources(tool_name, tool_version, subtool_name, base_dir=self.test_content_dir)
        command_line_tool = load_document(str(tool_sources['cwl']))
        with NamedTemporaryFile() as tf:
            command_line_tool.dump_cwl(tf.name)
            assert True
        return

    def test_dump_to_string(self):
        tool_name = 'cat'
        tool_version = '8.x'
        subtool_name = repo_config.main_tool_subtool_name
        tool_sources = get_tool_sources(tool_name, tool_version, subtool_name, base_dir=self.test_content_dir)
        command_line_tool = load_document(str(tool_sources['cwl']))
        cwl_string = command_line_tool.dump_cwl_str()
        return


class TestDumpCwlWorkflow(TestBase):
    pass


from ruamel.yaml import safe_load
from capanno_utils.classes.metadata.script_metadata import ScriptMetadata
from capanno_utils.classes.metadata.tool_metadata import ParentToolMetadata, SubtoolMetadata
from capanno_utils.classes.metadata.workflow_metadata import WorkflowMetadata
from capanno_utils.helpers.dict_tools import no_clobber_update
from capanno_utils.helpers.file_management import dump_dict_to_yaml_output
from capanno_utils.helpers.get_paths import *


# Todo before stable release: update function names to be consistent.

def make_tool_identifiers_list(base_dir=None):
    tools_map = make_tools_map_dict(base_dir=base_dir)
    tool_identifiers = list(tools_map.keys())
    return tool_identifiers

def make_tools_index(base_dir, index_path=tool_index_path):
    tool_identifiers = make_tool_identifiers_list(base_dir=base_dir)
    root_repo_path = Path(base_dir)
    identifier_index_path = root_repo_path / identifier_index_dir
    if not identifier_index_path.exists():
        identifier_index_path.mkdir()
    index_path = Path(base_dir) / index_path
    with index_path.open('w') as index_file:
        index_file.writelines(f"{identifier}\n" for identifier in tool_identifiers)
    return index_path.resolve()


def make_tools_map_dict(base_dir=None, specify_exists=False):
    tools_dir = get_root_tools_dir(base_dir=base_dir)
    tools_map = {}

    for tool_dir in tools_dir.iterdir():
        if tool_dir.name == '.DS_Store':
            continue
        assert tool_dir.is_dir()
        for version_dir in tool_dir.iterdir():
            if version_dir.name == '.DS_Store':
                continue
            assert version_dir.is_dir()
            tool_map = make_tool_version_dir_map(tool_dir.name, version_dir.name, base_dir=base_dir, specify_exists=specify_exists)
            no_clobber_update(tools_map, tool_map)
    return tools_map

def make_tools_map(outfile_path=None, base_dir=None, specify_exists=False):
    """
    Make a yaml file that specifies paths and attributes of tools in tool_dir.
    :param tool_dir (Path): Path of directory that contains tools.
    :param outfile_name(str): name of file that will be made.
    :return:
        None
    """

    tools_map = make_tools_map_dict(base_dir=base_dir, specify_exists=specify_exists)
    dump_dict_to_yaml_output(tools_map, output=outfile_path)
    return outfile_path


def make_main_tool_map(tool_name, base_dir=None):
    """
    Make a yaml file that specifies paths and attributes of tools in a single tool directory. If outfile is provided, dump contents to outfile.

    """
    main_tool_map = {}
    tool_dir = get_main_tool_dir(tool_name, base_dir=base_dir)
    for version_dir in tool_dir.iterdir():
        no_clobber_update(main_tool_map, make_tool_version_dir_map(tool_name, version_dir.name, base_dir=base_dir))
    return main_tool_map


def make_tool_version_dir_map(tool_name, tool_version, base_dir=None, specify_exists=False):
    tool_version_map = {}

    tool_version_dir = get_tool_version_dir(tool_name, tool_version, base_dir=base_dir)
    parent_metadata_path = get_tool_metadata(tool_name, tool_version, parent=True, base_dir=base_dir)
    parent_metadata = ParentToolMetadata.load_from_file(parent_metadata_path, check_index=False)
    parent_rel_path = get_relative_path(parent_metadata_path, base_path=base_dir)
    tool_version_map[parent_metadata.identifier] = {'metadataPath': str(parent_rel_path),
                                                    'metadataStatus': parent_metadata.metadataStatus,
                                                    'name': parent_metadata.name,
                                                    'versionName': parent_metadata.softwareVersion.versionName,
                                                    'type': 'parent'}
    for subtool_dir in tool_version_dir.iterdir():
        if subtool_dir.name in ['.DS_Store', 'common']:
            continue
        assert subtool_dir.is_dir(), subtool_dir.name+" is not a directory. You likely have an extra file."
        tool_name_length = len(
            tool_name)  # Use to get directory name string after 'tool_name'. In case there are underscores in tool_name.
        subtool_name = subtool_dir.name[tool_name_length + 1:]
        if subtool_name == '':
            subtool_name = None
        subdir_map = make_subtool_map(tool_name, tool_version, subtool_name, base_dir=base_dir, specify_exists=specify_exists)
        no_clobber_update(tool_version_map, subdir_map)
    return tool_version_map


def make_subtool_map(tool_name, tool_version, subtool_name, base_dir=None, specify_exists=False):
    subtool_metadata_path = get_tool_metadata(tool_name, tool_version, subtool_name=subtool_name, parent=False,
                                              base_dir=base_dir)
    subtool_metadata = SubtoolMetadata.load_from_file(subtool_metadata_path, check_index=False)
    subdir_map = {}
    subtool_rel_path = get_relative_path(subtool_metadata_path, base_path=base_dir)
    if specify_exists:
        files_exist = check_for_workflow_language_files(subtool_metadata_path)
        subdir_map[subtool_metadata.identifier] = {'metadataPath': str(subtool_rel_path),
                                               'name': subtool_metadata.name,
                                               'metadataStatus': subtool_metadata.metadataStatus,
                                               'cwlStatus': subtool_metadata.cwlStatus,
                                               'cwlExists': files_exist['cwl'],
                                               'nextflowStatus': subtool_metadata.nextflowStatus,
                                               'nextflowExists': files_exist['nextflow'],
                                               'snakemakeStatus': subtool_metadata.snakemakeStatus,
                                               'snakemakeExists': files_exist['snakemake'],
                                               'wdlStatus': subtool_metadata.wdlStatus,
                                               'wdlExists': files_exist['wdl'],
                                               'type': 'subtool'}
    else:
        subdir_map[subtool_metadata.identifier] = {'metadataPath': str(subtool_rel_path),
                                               'name': subtool_metadata.name,
                                               'metadataStatus': subtool_metadata.metadataStatus,
                                               'cwlStatus': subtool_metadata.cwlStatus,
                                               'nextflowStatus': subtool_metadata.nextflowStatus,
                                               'snakemakeStatus': subtool_metadata.snakemakeStatus,
                                               'wdlStatus': subtool_metadata.wdlStatus,
                                               'type': 'subtool'}
    return subdir_map



def make_tool_common_dir_map(tool_name, tool_version, base_dir):
    common_metadata_path = get_tool_common_dir(tool_name, tool_version, base_dir=base_dir) / common_tool_metadata_name
    common_metadata = ParentToolMetadata.load_from_file(common_metadata_path, check_index=False)
    common_dir_map = {}
    common_dir_map[common_metadata.identifier] = {'metadataPath': str(common_metadata_path),
                                                  'metadataStatus': common_metadata.metadataStatus,
                                                  'name': common_metadata.name,
                                                  'versionName': common_metadata.softwareVersion.versionName,
                                                  'type': 'parent'}
    return common_dir_map



def check_for_workflow_language_files(subtool_metadata_path):
    workflow_language_file_path_dict = get_tool_sources_from_metadata_path(subtool_metadata_path)
    files_exist = {}
    for source_type, source_path in workflow_language_file_path_dict.items():
        files_exist[source_type] = source_path.exists()
    return files_exist







def make_scripts_map_dict(base_dir=None):
    scripts_dir = get_root_scripts_dir(base_dir=base_dir)

    scripts_map = {}
    for group_dir in scripts_dir.iterdir():
        group_script_map = make_group_script_map(group_dir.name, base_dir=base_dir)
        no_clobber_update(scripts_map, group_script_map)
    return scripts_map

def make_script_maps(outfile_path, base_dir=None):
    scripts_map = make_scripts_map_dict(base_dir=base_dir)
    dump_dict_to_yaml_output(scripts_map, output=outfile_path)
    return


def make_group_script_map(group_name, base_dir=None):
    group_script_map = {}
    script_group_dir = get_script_group_dir(group_name, base_dir=base_dir)
    for project_dir in script_group_dir.iterdir():
        script_project_map = make_project_script_map(group_name, project_dir.name, base_dir=base_dir)
        no_clobber_update(group_script_map, script_project_map)
    return group_script_map


def make_project_script_map(group_name, project_name, base_dir=None):
    script_project_map = {}
    script_project_dir = get_script_project_dir(group_name, project_name, base_dir=base_dir)
    for version_dir in script_project_dir.iterdir():
        version_map = make_script_version_map(group_name, project_name, version_dir.name, base_dir=base_dir)
        no_clobber_update(script_project_map, version_map)
    return script_project_map


def make_script_version_map(group_name, project_name, version_name, base_dir=None):
    script_version_map = {}
    script_version_dir = get_script_version_dir(group_name, project_name, version_name, base_dir=base_dir)
    for script_dir in script_version_dir.iterdir():
        if script_dir.name == 'common':
            continue
        script_map = make_script_map(group_name, project_name, version_name, script_dir.name, base_dir=base_dir)
        no_clobber_update(script_version_map, script_map)
    return script_version_map


def make_script_map(group_name, project_name, version_name, script_name, base_dir=None):
    script_map = {}
    if script_name == 'common':  # script_name would also be common
        raise ValueError(f"Should not pass {script_name} to make map.")
    else:
        script_cwl_path = get_cwl_script(group_name, project_name, version_name, script_name, base_dir=base_dir)
        script_rel_path = get_relative_path(script_cwl_path, base_path=base_dir)
        metadata_path = get_metadata_path(script_cwl_path)
        script_metadata = ScriptMetadata.load_from_file(metadata_path)
        script_map[script_metadata.identifier] = {'path': str(script_rel_path), 'name': script_metadata.name,
                                                  'versionName': script_metadata.softwareVersion.versionName,
                                                  'metadataStatus': script_metadata.metadataStatus,
                                                  'cwlStatus': script_metadata.cwlStatus}
    return script_map

# Workflow maps

def make_workflow_maps_dict(base_dir=None):
    workflows_dir = get_workflows_root_dir(base_dir=base_dir)
    master_workflow_map = {}
    for group_dir in workflows_dir.iterdir():
        for project_dir in group_dir.iterdir():
            for version_dir in project_dir.iterdir():
                workflow_dict = make_workflow_map(group_dir.name, project_dir.name, version_dir.name,
                                                          base_dir=base_dir)
                no_clobber_update(master_workflow_map, workflow_dict)
    return master_workflow_map

def make_workflow_maps(outfile_name='workflow-maps', base_dir=None):
    master_workflow_map = make_workflow_maps_dict(base_dir=base_dir)
    dump_dict_to_yaml_output(master_workflow_map, output=outfile_name)
    return

def make_group_workflow_map(group_name, base_dir=None):
    group_workflow_map = {}
    workflow_group_dir = get_workflow_group_dir(group_name, base_dir)
    for project_dir in workflow_group_dir.iterdir():
        workflow_project_map = make_project_workflow_map(group_name, project_dir.name, base_dir)
        no_clobber_update(group_workflow_map, workflow_project_map)
    return group_workflow_map

def make_project_workflow_map(group_name, project_name, base_dir=None):
    workflow_project_map = {}
    workflow_project_dir = get_workflow_project_dir(group_name, project_name, base_dir)
    for version_dir in workflow_project_dir.iterdir():
        version_map = make_workflow_map(group_name, project_name, version_dir.name, project_name, base_dir)
        no_clobber_update(workflow_project_map, version_map)
    return workflow_project_map

def make_version_workflow_map(group_name, project_name, version_name, base_dir=None):
    workflow_version_map = make_workflow_map(group_name, project_name, version_name, project_name, base_dir)
    return workflow_version_map

def make_workflow_map(group_name, project_name, version, base_dir=None):
    workflow_map = {}
    workflow_metadata_path = get_workflow_metadata(group_name, project_name, version, base_dir=base_dir)
    workflow_metadata = WorkflowMetadata.load_from_file(workflow_metadata_path)
    workflow_metadata_rel_path = get_relative_path(workflow_metadata_path, base_path=base_dir)
    workflow_map[workflow_metadata.identifier] = {'metadataPath': str(workflow_metadata_rel_path), 'name': workflow_metadata.name,
                                                  'metadataStatus': workflow_metadata.metadataStatus,
                                                  'workflowLanguage': workflow_metadata.workflowLanguage,
                                                  'workflowStatus': workflow_metadata.workflowStatus,
                                                  'workflowPath': workflow_metadata.workflowFile,
                                                  'versionName': workflow_metadata.softwareVersion.versionName,
                                                }
    return workflow_map


def combine_yaml_files_into_dict(file_path, *file_paths):
    # make a 'master map' that contains all file
    combined_dict = {}
    with file_path.open('r') as f:
        no_clobber_update(combined_dict, safe_load(f))
    for file in file_paths:
        with file.open('r') as f:
            no_clobber_update(combined_dict, safe_load(f))
    return combined_dict


def make_master_map_dict(base_dir=None, specify_exists=False):
    master_map = {}
    master_map.update(make_tools_map_dict(base_dir=base_dir, specify_exists=specify_exists))
    master_map.update(make_scripts_map_dict(base_dir=base_dir))
    master_map.update(make_workflow_maps_dict(base_dir=base_dir))
    return master_map

def make_master_map(file_name, *file_names, outfile_name="master_map"):
    file_path = config[os.environ['CONFIG_KEY']]['content_maps_dir'] / f"{file_name}.yaml"

    raise NotImplementedError


def get_tool_map():
    raise NotImplementedError


def add_to_map(path):
    """
    Add tools scripts and workflows with versions >= 1.0 to content_maps tool_maps, script_maps, or 'workflow_maps'.
    These maps provide an accessible way to work with content based on their identifiers.
    :param path:
    :return:
    """
    raise NotImplementedError


def get_tool_args_from_identifier(identifier, base_dir=None):
    base_dir = get_base_dir(base_dir)
    instance_hash = None
    if tool_instance_identifier_pattern.match(identifier):
        identifier, instance_hash = identifier[:-4], identifier[-4:]
    if parent_tool_identifier_pattern.match(identifier) | subtool_identifier_pattern.match(identifier):
        tools_map = make_tools_map(base_dir=base_dir)
        tool_path = tools_map[identifier]['metadataPath']
        tool_args = get_tool_args_from_path(tool_path)
    else:
        raise ValueError()
    return tool_args, instance_hash


#!/usr/bin/env python3

from datetime import date
from capanno_utils.classes.metadata.tool_metadata import ParentToolMetadata, SubtoolMetadata
from capanno_utils.initialize_wf_files import initialize_tool_wf_file_tool
from capanno_utils.classes.schema_salad.schema_salad import InputsSchema
from capanno_utils.content_maps import make_tools_index
from capanno_utils.helpers.get_paths import *
import logging, sys

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def add_tool(tool_name, version_name, subtool_names=None, biotools_id=None, has_primary=False, root_repo_path=Path.cwd(), init_cwl=False, init_wdl=False, init_sm=False, init_nf=False, no_clobber=False, refresh_index=True):
    """
    Make the correct directory structure for adding a new command line tool. Optionally, create initialized wf language
    and metadata files. Run from tools directory.
    :param tool_name(str): Name of the tool
    :param version_name(str): version of the tool
    :param subtool_names(list(str)): list of subtool names if the tool is broken into multiple subtools.
    :return: None
    """
    version_name = str(version_name) # In case ArgumentParser is bypassed.
    if subtool_names:
        if isinstance(subtool_names, str):
            subtool_names = [subtool_names]
    if has_primary:  # Need to append __main__ onto subtools.
        if subtool_names:
            subtool_names.append(main_tool_subtool_name)
        else: # __main__ is only subtool specified.
            subtool_names = [main_tool_subtool_name]
            if isinstance(init_cwl, str):
                init_cwl = {main_tool_subtool_name: init_cwl}

    if refresh_index:
        make_tools_index(base_dir=root_repo_path)
    common_dir = get_tool_common_dir(tool_name, version_name, base_dir=root_repo_path)
    if no_clobber and common_dir.exists():
        logger.debug("Skipping tool directory setup")
        return
    common_dir.mkdir(parents=True, exist_ok=not no_clobber)
    if biotools_id:
        # tool_name will be ignored.
        parent_metadata = ParentToolMetadata.create_from_biotools(biotools_id, version_name, subtool_names, name=tool_name, root_repo_path=root_repo_path)
        # can't find it in biotools
        if parent_metadata is None:
            parent_metadata = ParentToolMetadata(name=tool_name,
                                                 softwareVersion={'versionName': version_name, 'includedVersions': []},
                                                 featureList=subtool_names, check_index=True, _in_index=False)
    else:
        parent_metadata = ParentToolMetadata(name=tool_name, softwareVersion={'versionName': version_name, 'includedVersions': []}, featureList=subtool_names, root_repo_path=root_repo_path, check_index=True, _in_index=False)
    if parent_metadata.featureList:
        for subtool in parent_metadata.featureList:
            subtool_obj = parent_metadata.make_subtool_metadata(subtool_name=subtool, root_repo_path=root_repo_path, check_index=True)
            subtool_obj.mk_file()
            subtool_dir = get_tool_dir(tool_name, version_name, subtool, base_dir=root_repo_path)
            instances_dir = subtool_dir / 'instances'
            instances_dir.mkdir()
            git_keep_file = instances_dir / '.gitkeep'
            git_keep_file.touch()
            if isinstance(init_cwl, dict):
                init_cwl_ = init_cwl.get(subtool, False)  # Not initialized if not specified.
            else:
                init_cwl_ = init_cwl
            initialize_tool_wf_file_tool(tool_name, version_name, subtool, init_cwl=init_cwl_, init_wdl=init_wdl, init_sm=init_sm, init_nf=init_nf, base_dir=root_repo_path)
    parent_metadata.mk_file(root_repo_path)
    return


def add_subtool(tool_name, tool_version, subtool_name, root_repo_path=Path.cwd(), update_featureList=False, init_cwl=False, init_wdl=False, init_sm=False, init_nf=False, no_clobber=False, refresh_index=True):
    """
    Add subtool to already existing ToolLibrary (ParentTool file already exists)
    :param tool_name(str):
    :param version_name (str):
    :param subtool_name (str):
    :param root_repo_path (Path): The local directory under which this subtool should be added
    :param update_featureList (Bool): If True, subtool does not need to be in ParentTool featureList and ParentTool will be updated. Will throw error if False and subtool is not in ParentTool featureList.
    :param init_cwl:
    :return:
    """
    if refresh_index:
        make_tools_index(base_dir=root_repo_path)
    subtool_kwargs = {'extra': {}}  # initialize to add any additional information about the subtool.
    parent_path = get_tool_metadata(tool_name, tool_version, parent=True, base_dir=root_repo_path)
    parent_meta = ParentToolMetadata.load_from_file(parent_path, root_repo_path=root_repo_path, _in_index=True)  # When adding a subtool, the ParentTool should already be present and in the index.

    subtool_dir = get_tool_dir(tool_name, tool_version, subtool_name, base_dir=root_repo_path)
    if no_clobber and subtool_dir.exists():
        return
    if update_featureList:
        if parent_meta.featureList is None:
            parent_meta.featureList = [subtool_name]
        else:
            if not subtool_name in parent_meta.featureList:
                parent_meta.featureList.append(subtool_name)
        parent_meta.mk_file(base_dir=root_repo_path, update_index=False)  # Remake the file. Needs to be remade if updated. Identifier will already be in index. No place to update the identifier in this function.
    if not isinstance(init_cwl, bool):  # initialized from a url.
        subtool_kwargs['extra'].update({'cwlDocument': {'isBasedOn': init_cwl, 'dateCreated': str(date.today())}})
    if not isinstance(init_wdl, bool):  # initialized from a url.
        subtool_kwargs['extra'].update({'wdlDocument': {'isBasedOn': init_wdl, 'dateCreated': str(date.today())}})
    if not isinstance(init_sm, bool):  # initialized from a url.
        subtool_kwargs['extra'].update({'snakemakeDocument': {'isBasedOn': init_sm, 'dateCreated': str(date.today())}})
    if not isinstance(init_nf, bool):  # initialized from a url.
        subtool_kwargs['extra'].update({'nextflowDocument': {'isBasedOn': init_nf, 'dateCreated': str(date.today())}})
    subtool_meta = parent_meta.make_subtool_metadata(subtool_name, root_repo_path=root_repo_path, check_index=True, **subtool_kwargs)
    subtool_meta.mk_file()
    instances_dir = subtool_dir / 'instances'
    instances_dir.mkdir()
    git_keep_file = instances_dir / '.gitkeep'
    git_keep_file.touch()
    initialize_tool_wf_file_tool(tool_name, tool_version, subtool_name, init_cwl=init_cwl, init_wdl=init_wdl, init_sm=init_sm, init_nf=init_nf, base_dir=root_repo_path)
    return


def add_tool_instance(tool_name, tool_version, subtool_name, init_job_file=True, root_repo_path=Path.cwd()):
    subtool_path = get_tool_metadata(tool_name, tool_version, subtool_name, base_dir=root_repo_path)
    subtool_metadata = SubtoolMetadata.load_from_file(subtool_path, root_repo_path=root_repo_path, _in_index=False)
    instance_meta = subtool_metadata.mk_instance()
    instance_metadata_path = instance_meta.mk_file(base_dir=root_repo_path)
    input_hash = instance_meta.identifier[-4:]
    job_file_path = None
    if init_job_file:
        tool_sources = get_tool_sources(tool_name, tool_version, subtool_name, base_dir=root_repo_path)
        job_file_path = get_tool_instance_path(tool_name, tool_version, input_hash=input_hash, subtool_name=subtool_name, base_dir=root_repo_path)
        tool_inputs_schema = InputsSchema(tool_sources['cwl'])
        tool_inputs_schema.make_template_file(job_file_path)
    return instance_metadata_path, job_file_path



from pathlib import Path
from capanno_utils.make_content_maps import make_tools_map_dict
from capanno_utils.classes.metadata.tool_metadata import SubtoolMetadata
from capanno_utils.helpers.get_paths import *

unpublished_statuses = ('Draft', 'Incomplete')

def update_all_tools_to_released(base_dir=None, update_dir=None):
    tool_map = make_tools_map_dict(base_dir=base_dir, specify_exists=True)

    for identifier, values in tool_map.items():
        if values['type'] == 'subtool':
            update_existing_tool_wrapper_status(values, base_dir=base_dir, update_dir=update_dir)

    return


def update_existing_tool_wrapper_status(tool_map_values, base_dir=None, update_dir=None):
    base_dir = get_base_dir(base_dir=base_dir)
    metadata_path = Path(base_dir) / tool_map_values['metadataPath']
    subtool_metadata = SubtoolMetadata.load_from_file(metadata_path)
    if tool_map_values['cwlStatus'] in unpublished_statuses and tool_map_values['cwlExists'] == True:
        subtool_metadata.cwlStatus = 'Released'
    if tool_map_values['wdlStatus'] in unpublished_statuses and tool_map_values['wdlExists'] == True:
        subtool_metadata.wdlStatus = 'Released'
    if tool_map_values['nextflowStatus'] in unpublished_statuses and tool_map_values['nextflowExists'] == True:
        subtool_metadata.nextflowStatus = 'Released'
    if tool_map_values['snakemakeStatus'] in unpublished_statuses and tool_map_values['snakemakeExists'] == True:
        subtool_metadata.snakemakeStatus = 'Released'
    if update_dir == None or base_dir == update_dir:  # Need to overwrite the existing file.
        subtool_metadata.mk_file()
    else:
        # updated_path = Path(update_dir) / tool_map_values['metadataPath']
        subtool_metadata.root_repo_path = update_dir
        subtool_metadata.mk_file()

    return



# Someday these files will be stored in the content repo themselves.

import os
import re
import tempfile
from pathlib import Path

test_rel_content_path = Path.cwd() / 'tests' / 'test_files' / 'capanno'
if os.environ.get('TRAVIS_BUILD_DIR'):
    travis_root = Path(os.environ['TRAVIS_BUILD_DIR']) / 'tests' / 'test_files' / 'capanno'
else:
    travis_root = Path.cwd()

content_base_path = Path.cwd()

content_maps_dir_name = 'content_maps'

main_tool_subtool_name = '__main__'  # store here so can change easily. This is the 'name' of the main tool (categorized as a subtool).

instance_file_pattern = re.compile(r'[0-9a-f]{4}\.ya?ml')

instance_metadata_file_pattern = re.compile(r'[0-9a-f]{4}-metadata\.ya?ml')

metadata_file_pattern = re.compile(r'[.0-9A-Za-z_-]+-metadata\.ya?ml')

parent_tool_identifier_pattern = re.compile(r'TL_[0-9a-f]{6}\.[0-9a-f]{2}$')

subtool_identifier_pattern = re.compile(r'TL_[0-9a-f]{6}_[0-9a-f]{2}\.[0-9a-f]{2}$')

tool_instance_identifier_pattern = re.compile(r'TL_[0-9a-f]{6}_[0-9a-f]{2}\.[0-9a-f]{2}\.[0-9a-f]{4}$')

workflow_identifier_pattern = re.compile(r'WF_[0-9a-f]{6}\.[0-9a-f]{2}$')

workflow_instance_identifier_pattern = re.compile(r'WF_[0-9a-f]{6}\.[0-9a-f]{2}\.[0-9a-f]{4}$')

common_dir_name = 'common'

common_tool_metadata_name = 'common-metadata.yaml'

instances_dir_name = 'instances'

content_repo_name = 'capanno'

tools_dir_name = 'tools'

scripts_dir_name = 'scripts'

workflows_dir_name = 'workflows'

tool_identifier_prefix = 'TL'

script_identifier_prefix = 'ST'

worklfow_identifier_prefix = 'WF'

tools_map_name = '.tools_map.yaml'

scripts_maps_name = '.scripts_map.yaml'

workflow_maps_name = 'workflows_map.yaml'  # why are the others hidden and this one isn't?

identifier_index_dir = Path('.cache')

tool_index_file_name = 'tools_index'

tool_index_path = identifier_index_dir / tool_index_file_name


def make_config_dict(base_path):
    base_path = Path(base_path)  # Make sure any string values are turned into Path objects.
    config_dict = {
        'base_path': base_path,
        'content_maps_dir': base_path / content_maps_dir_name,
        'temp_dir': tempfile.TemporaryDirectory(prefix=f"{content_repo_name}_Test_"),
    }
    return config_dict


# Default values for using tool in repos.
_default = make_config_dict(content_base_path)

# Test values for running unit tests.
_test = make_config_dict(test_rel_content_path)

_travis = make_config_dict(travis_root)

config = {'TEST': _test, 'DEFAULT': _default, 'TRAVIS': _travis}


def make_full_config_dict():
    config_dict = _default
    constants = dict([
        ('content_maps_dir_name', 'content_maps'),
        ('main_tool_subtool_name', '__main__'),
        ('common_dir_name', 'common'),
        ('common_tool_metadata_name', 'common-metadata.yaml'),
        ('instances_dir_name', 'instances'),
        ('content_repo_name', 'capanno'),
        ('tools_dir_name', 'tools'),
        ('scripts_dir_name', 'scripts'),
        ('workflows_dir_name', 'workflows'),
        ('scripts_maps_name', '.scripts_map.yaml'),
        ('workflow_maps_name', 'workflows_map.yaml'),
        ('tool_index_file_name', 'tools_index'),
    ])
    config_dict.update(constants)
    return config_dict

config_dict = make_full_config_dict()



import tempfile
import shutil
from unittest import TestCase
import os
from pathlib import Path
from capanno_utils import repo_config
from capanno_utils.helpers.get_paths import *
from capanno_utils.content_maps import make_tools_map, make_workflow_maps, make_script_maps, make_tools_index

test_constants = {'script_group1': 'ENCODE-DCC', 'script_version1': '1.1.x', 'script_project1': 'atac-seq-pipeline',
                  'test_software_version': {'versionName': 'test_version'}}


class TestBase(TestCase):
    projects_path = Path(__file__).parents[2]
    src_content_dir = projects_path / repo_config.content_repo_name
    test_files_dir = Path(__file__).parent / 'test_files'
    test_content_dir = Path(__file__).parent / 'test_files' / repo_config.content_repo_name
    invalid_content_dir = Path(__file__).parent / 'test_files' / 'invalid_content'  # Will copy data here, then modify to make invalid.

    def get_content_map_paths(self):
        return {'tool_maps': repo_config.config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / repo_config.tools_map_name,
                'script_maps': repo_config.config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / 'script-maps.yaml',
                'workflow_maps': repo_config.config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / 'workflow-maps.yaml'}


    def update_tool_maps(self):
        outfile_path = self.get_content_map_paths()['tool_maps']
        make_tools_map(outfile_path)

    def update_script_maps(self):
        outfile_path = self.get_content_map_paths()['script_maps']
        make_script_maps(outfile_path)

    def update_workflow_maps(self):
        outfile_path = self.get_content_map_paths()['workflow_maps']
        make_workflow_maps(outfile_path)

    def make_empty_tools_index(self, root_repo_dir):
        root_repo_dir = Path(root_repo_dir)
        tools_index_dir = root_repo_dir / repo_config.identifier_index_dir
        tools_index_dir.mkdir()
        tool_index_path = root_repo_dir / repo_config.tool_index_path
        tool_index_path.touch()
        return

    def make_empty_tools_dir(self, root_repo_path):
        output_dir = Path(root_repo_path)
        tools_dir = output_dir / repo_config.tools_dir_name
        tools_dir.mkdir(parents=True)
        return

    def copy_all_content_repo_tools(self, dest):
        tool_src = self.src_content_dir / repo_config.tools_dir_name
        tool_dest = Path(dest) / repo_config.tools_dir_name
        shutil.copytree(tool_src, tool_dest)
        make_tools_index(base_dir=dest)
        return



    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()

    def tearDown(self) -> None:
        pass




import os
from tests.test_base import TestBase, test_constants
from capanno_utils.helpers.get_paths import *
    # get_types_from_path, get_script_metadata, get_script_instance_path, \
    # get_cwl_script, get_cwl_tool, get_tool_common_dir, main_tool_subtool_name, get_tool_metadata, \
    # get_tool_instance_path, get_tool_instance_metadata_path, get_cwl_workflow, get_workflow_metadata, \
    # get_workflow_instance_path, get_workflow_instance_metadata


class TestGetTypesFromPath(TestBase):

    def test_get_type_from_script_paths(self):

        # Script cwl path
        script_cwl_path = get_cwl_script(test_constants['script_group1'], test_constants['script_project1'],
                                         test_constants['script_version1'], 'encode_ataqc')
        cwl_script_type_tuple = get_types_from_path(str(script_cwl_path))
        self.assertEqual(cwl_script_type_tuple, ('script', 'cwl'))
        script_metadata_path = get_script_metadata(test_constants['script_group1'], test_constants['script_project1'],
                                         test_constants['script_version1'], 'encode_ataqc')

        # Script metadata path
        script_metadata_type_tuple = get_types_from_path(str(script_metadata_path))
        self.assertEqual(script_metadata_type_tuple, ('script', 'metadata'))

        # Script instance/metadata paths. ToDo Need to have a script instance to test.

        # script_instance_path = get_script_instance_path(test_constants['script_group1'], test_constants['script_project1'],
        #                                  test_constants['script_version1'], 'encode_ataqc', '1234')
        # script_instance_type_tuple = get_types_from_path(script_instance_path)

        return

    def test_get_type_from_tool_paths(self):

        tool_name = 'gawk'
        tool_version = '4.1.x'
        subtool_name = main_tool_subtool_name
        instance_hash = '1c10'

        common_metadata_path = get_tool_common_dir(tool_name, tool_version) / 'common-metadata.yaml'
        common_tool_type_tuple = get_types_from_path(common_metadata_path)
        self.assertEqual(common_tool_type_tuple, ('tool', 'common_metadata'))

        tool_sources = get_tool_sources(tool_name, tool_version, subtool_name)
        cwl_tool_type_tuple = get_types_from_path(tool_sources['cwl'])
        self.assertEqual(cwl_tool_type_tuple, ('tool', 'cwl'))

        tool_metadata_path = get_tool_metadata(tool_name, tool_version)
        tool_metadata_type_tuple = get_types_from_path(tool_metadata_path)
        self.assertEqual(tool_metadata_type_tuple, ('tool', 'metadata'))

        tool_instance_path = get_tool_instance_path(tool_name, tool_version, instance_hash)
        tool_instance_type_tuple = get_types_from_path(tool_instance_path)
        self.assertEqual(tool_instance_type_tuple, ('tool', 'instance'))

        tool_instance_metadata_path = get_tool_instance_metadata_path(tool_name, tool_version, instance_hash)
        tool_instance_metadata_type_tuple = get_types_from_path(tool_instance_metadata_path)
        self.assertEqual(tool_instance_metadata_type_tuple, ('tool', 'instance_metadata'))

        return

    def test_get_type_from_workflow_path(self):

        group_name = 'example_workflows'
        project_name = 'cat_sort'
        workflow_version = 'master'
        workflow_name = 'cat_sort'
        input_hash = 'ecd8'

        workflow_cwl_path = get_workflow_sources(group_name, project_name, workflow_version, workflow_name)['cwl']
        cwl_workflow_type_tuple = get_types_from_path(workflow_cwl_path)
        self.assertEqual(cwl_workflow_type_tuple, ('workflow', 'cwl'))

        workflow_metadata_path = get_workflow_metadata(group_name, project_name, workflow_version)
        workflow_metadata_type_tuple = get_types_from_path(workflow_metadata_path)
        self.assertEqual(workflow_metadata_type_tuple, ('workflow', 'metadata'))

        workflow_instance_path = get_workflow_instance_path(group_name, project_name, workflow_version, input_hash)
        workflow_instance_type_tuple = get_types_from_path(workflow_instance_path)
        self.assertEqual(workflow_instance_type_tuple, ('workflow', 'instance'))

        workflow_instance_metadata_path = get_workflow_instance_metadata(group_name, project_name, workflow_version, input_hash)
        workflow_instance_metadata_type_tuple = get_types_from_path(workflow_instance_metadata_path)
        self.assertEqual(workflow_instance_metadata_type_tuple, ('workflow', 'instance_metadata'))
        return


    def test_get_type_from_tool_dirs(self):

        tool_name = 'gawk'
        tool_version = '4.1.x'
        subtool_name = main_tool_subtool_name  # Is used by default if not specified.
        cwl_root_repo_name = self.test_content_dir.stem

        root_tools_dir = get_root_tools_dir()
        root_tools_dir_type_tuple = get_types_from_path(root_tools_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(root_tools_dir_type_tuple, ('tool', 'base_dir'))


        tool_dir = get_main_tool_dir(tool_name)
        tool_dir_type_tuple = get_types_from_path(tool_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(tool_dir_type_tuple, ('tool', 'tool_dir'))

        tool_version_dir = get_tool_version_dir(tool_name, tool_version)
        tool_version_dir_type_tuple = get_types_from_path(tool_version_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(tool_version_dir_type_tuple, ('tool', 'version_dir'))

        tool_common_dir = get_tool_common_dir(tool_name, tool_version)
        tool_common_dir_type_tuple = get_types_from_path(tool_common_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(tool_common_dir_type_tuple, ('tool', 'common_dir'))

        subtool_dir = get_tool_dir(tool_name, tool_version)
        subtool_dir_type_tuple = get_types_from_path(subtool_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(subtool_dir_type_tuple, ('tool', 'subtool_dir'))

        tool_instance_dir = get_tool_instances_dir(tool_name, tool_version)
        tool_instance_dir_type_tuple = get_types_from_path(tool_instance_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(tool_instance_dir_type_tuple, ('tool', 'instance_dir'))
        return

    def test_get_type_from_script_dirs(self):

        group_name = 'ENCODE-DCC'
        project_name = 'atac-seq-pipeline'
        script_version = '1.1.x'
        script_name = 'encode_ataqc'
        cwl_root_repo_name = self.test_content_dir.stem

        root_scripts_dir = get_root_scripts_dir()
        root_scripts_dir_tuple_type = get_types_from_path(root_scripts_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(root_scripts_dir_tuple_type, ('script', 'base_dir'))

        script_group_dir = get_script_group_dir(group_name)
        script_group_dir_type_tuple = get_types_from_path(script_group_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(script_group_dir_type_tuple, ('script', 'group_dir'))

        script_project_dir = get_script_project_dir(group_name, project_name)
        script_project_dir_type_tuple = get_types_from_path(script_project_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(script_project_dir_type_tuple, ('script', 'project_dir'))

        script_version_dir = get_script_version_dir(group_name, project_name, script_version)
        script_version_type_tuple = get_types_from_path(script_version_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(script_version_type_tuple, ('script', 'version_dir'))

        script_dir = get_script_dir(group_name, project_name, script_version, script_name)
        script_dir_type_tuple = get_types_from_path(script_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(script_dir_type_tuple, ('script', 'script_dir'))

        script_instance_dir = get_script_instance_dir(group_name, project_name, script_version, script_name)
        script_instance_dir_tuple = get_types_from_path(script_instance_dir, root_repo_name=cwl_root_repo_name)
        self.assertEqual(script_instance_dir_tuple, ('script', 'instance_dir'))



        return





from pathlib import Path
import re
from collections import OrderedDict
from ruamel.yaml import safe_load
from capanno_utils.repo_config import *
from ...classes.metadata.common_functions import is_attr_empty, CommonPropsMixin, WorkflowLanguageStatusMixin, SoftwarePropsMixin, _mk_hashes
from ...classes.metadata.shared_properties import WebSite, CodeRepository, Person, Publication, Keyword, ParentScript, \
    Tool
from ...classes.metadata.metadata_base import MetadataBase


class ScriptMetadataBase(MetadataBase):
    """Factor stuff out to here if there is more than one ScriptMetadata class."""

    @property
    def parentScripts(self):
        return self._parentScripts

    @parentScripts.setter
    def parentScripts(self, parent_script_list):
        if parent_script_list:
            parent_scripts = []
            for parent_script in parent_script_list:
                if isinstance(parent_script, ParentScript):
                    parent_scripts.append(parent_script)
                else:
                    parent_scripts.append(ParentScript(**parent_script))
        else:
            parent_scripts = None
        self._parentScripts = parent_scripts

    @property
    def programmingLanguage(self):
        return self._programmingLanguage

    @programmingLanguage.setter
    def programmingLanguage(self,programming_language):
        if not isinstance(programming_language, (str, type(None))):
            raise ValueError(f"programmingLanguage must be set to a string. You provided {programming_language}")
        self._programmingLanguage = programming_language


    @property
    def tools(self):
        return self._tools

    @tools.setter
    def tools(self, tools_list):
        if tools_list:
            tools = []
            for tool in tools_list:
                if isinstance(tool, Tool):
                    tools.append(tool)
                else:
                    tools.append(Tool(**tool))
        else:
            tools = None
        self._tools = tools




class ScriptMetadata(CommonPropsMixin, WorkflowLanguageStatusMixin, SoftwarePropsMixin, ScriptMetadataBase):


    @staticmethod
    def _init_metadata():
        return OrderedDict([
            ('name', None),
            ('softwareVersion', None),
            ('identifier', None),
            ('current', False),
            ('metadataStatus', 'Incomplete'),
            ('cwlStatus', 'Incomplete'),
            ('nextflowStatus', 'Incomplete'),
            ('snakemakeStatus', 'Incomplete'),
            ('wdlStatus', 'Incomplete'),
            ('description', None),
            ('codeRepository', None),
            ('WebSite', None),
            ('license', None),
            ('contactPoint', None),
            ('publication', None),
            ('keywords', [Keyword()]),
            ('parentScripts', None),
            ('tools', None),
            ('alternateName', None),
            ('creator', None),
            ('programmingLanguage', None),
            ('datePublished', None),
            ('isCallable', True),  # specify whether the script is meant to be called. If False, just contains attribute_value for importing to other scripts.
            ('parentMetadata', None),
            ('_parentMetadata', None),  # Place to store list of parent ScriptMetadata objects. Different than parentScripts!!
            ('_primary_file_attrs', None),
            # List to store attributes that are not inherited from a parent metadata file or attribute_value.
        ])

    def __init__(self, file_path=Path.cwd(), **kwargs):
        """
        What it's gotta do: see if there's anything in parentMetadata. If so, load it into _parentMetadata. Check for values in kwargs.
        If value is there, leave it alone and put key in _primary_file_attrs. If it is not there, check for highest priority value in _parent_metadada and set to that value.
        The main question. How to do this with the base class init which only expects kwargs as arguments. What if value already set... Think this is taken care of.
        :param file_path:
        :param kwargs:
        """

        ignore_empties = kwargs.pop('ignore_empties', None)
        # Get parent metadata first.
        master_common_metadata = None  # Set to None so nothing happens if no common metadata.
        if kwargs.get('parentMetadata'):
            self.parentMetadata = kwargs['parentMetadata']
            self._load_common_metadata(file_path)  # Will set self._parentMetadata
            master_common_metadata = self._mk_master_common_metadata()

            # populate _primary_attrs and populate required values from common_metadata where required.
            self._primary_file_attrs = []
            for k, value in kwargs.items():
                if value:
                    self._primary_file_attrs.append(k)

        super().__init__(ignore_empties=ignore_empties, **kwargs)
        if master_common_metadata:
            self._update_attributes(master_common_metadata)

    def _check_identifier(self, identifier):
        if not identifier[:3] == f"{script_identifier_prefix}_":
            raise ValueError(f"Script identifiers must start with '{script_identifier_prefix}_' you provided {identifier}")
        else:
            hex_pattern = r'[0-9a-f]{6}\.[0-9a-f]{2}$'
            match_obj = re.match(hex_pattern, identifier[3:])
            if not match_obj:
                raise ValueError(f"Script identifier not formatted correctly: {identifier}")

        return identifier

    def _mk_identifier(self, start=0):
        if not (self.name and self.softwareVersion.versionName):
            raise ValueError(f"Name and softwareVersion must be provided to make an identifier.")
        name_hash, version_hash = _mk_hashes(self.name, self.softwareVersion.versionName)
        identifier = f"{script_identifier_prefix}_{name_hash[start:start + 6]}.{version_hash[:2]}"
        return identifier

    def _load_common_metadata(self, file_path):
        """
        :param file_path: path of the original file that specifies parent/commonMetadata. parentMetadata file paths are relative to this.
        :return:
        """
        self._parentMetadata = []  # List of parent CommonScriptMetadata objects.
        if self.parentMetadata:
            dirname = file_path.parent
            for rel_path in self.parentMetadata:
                full_path = dirname / rel_path
                with full_path.resolve().open('r') as f:
                    common_meta_dict = safe_load(f)
                self._parentMetadata.append(CommonScriptMetadata(**common_meta_dict))
        return

    @classmethod
    def load_from_file(cls, file_path, ignore_empties=False):
        file_path = Path(file_path)
        with file_path.open('r') as file:
            file_dict = safe_load(file)
        return cls(file_path=file_path, ignore_empties=ignore_empties, **file_dict)


    def _update_attributes(self, update_instance):
        """
        Update self with attribute values in update_instance for attributes in self that have not been set.
        :param update_instance:
        :return:
        """
        for attribute_name in self._init_metadata().keys():
            orig_attribute_value = getattr(self, attribute_name)  # See if it's already set on self. If it is, leave it alone.
            try:
                update_instance_attribute_value = getattr(update_instance, attribute_name)
            except AttributeError:
                continue  # Attribute doesn't exist for update_instance so can't be used to update. Move on.

            if orig_attribute_value:  # There's already something there, leave it alone. Might combine data at some point.
                continue
            if is_attr_empty(orig_attribute_value):
                assert True  # Nothing present initally
                if is_attr_empty(update_instance_attribute_value):
                    continue  # Nothing to do here. Move along.
                else:
                    setattr(self, attribute_name, update_instance_attribute_value)
        return

    def _mk_master_common_metadata(self):
        # Combine all common metadata files into a master CommonScriptMetadata instance where lower priority metadata is overwritten by higher priority metadata.
        # This will be used to update the ScriptMetadata instance.
        len_parent_metadata = len(self._parentMetadata)
        if len_parent_metadata == 0:
            master_common_metadata =  None
        elif len_parent_metadata == 1:
            master_common_metadata = self._parentMetadata[0]
        else:
            master_common_metadata = self._parentMetadata[0]    # initialize
            for common_script_metadata in self._parentMetadata[1:]:  # skip first entry since it is used to initialize.
                master_common_metadata.update_from_other_instance(common_script_metadata)
        return master_common_metadata


    def mk_file(self, file_path):
        return super().mk_file(file_path, keys=self._primary_file_attrs)


    def mk_completed_file(self, file_path):
        # substitute in parent metadata fields for fields not specified by the script's own metadata.
        return super().mk_file(file_path)

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, identifier=None, **kwargs):
        if identifier:
            identifier = self._check_identifier(identifier)
        else:
            identifier = self._mk_identifier(**kwargs)
        self._identifier = identifier


    @property
    def isCallable(self):
        return self._is_callable

    @isCallable.setter
    def isCallable(self, callable_value):
        if not isinstance(callable_value, (bool, type(None))):
            assert True
            raise TypeError(f"callable must be a boolean, you provided {callable_value}")
        self._is_callable = callable_value



class CommonScriptMetadata(CommonPropsMixin, SoftwarePropsMixin, ScriptMetadataBase):


    @staticmethod
    def _init_metadata():
        # Only attributes which can be common to multiple scripts. Skips validation of codeRepository
        return OrderedDict([
            ('softwareVersion', None),
            ('description', None),
            ('codeRepository', None),
            ('WebSite', None),
            ('license', None),
            ('contactPoint', None),
            ('publication', None),
            ('keywords', None),
            ('creator', None),
            ('programmingLanguage', None),
            ('datePublished', None),
            ('parentScripts', None),
            ('tools', None),
        ])

    def update_from_other_instance(self, common_script_metadata_instance):
        # Update self from common_script_metadata_instance
        for attribute in self._init_metadata():
            if is_attr_empty(getattr(self, attribute)):
                update_value = getattr(common_script_metadata_instance, attribute)
                if is_attr_empty(update_value):
                    continue
                else:
                    setattr(self, attribute, update_value)
            else:
                continue
        return

    # override softwareVersion so it is not required for CommonScriptMetadata
    @property
    def softwareVersion(self):
        return super().softwareVersion

    @softwareVersion.setter
    def softwareVersion(self, software_version_info):
        if not software_version_info:
            self._softwareVersion = None  # Allows softwar
        else:
            super(CommonScriptMetadata, type(self)).softwareVersion.fset(self, software_version_info)

    @classmethod
    def load_from_file(cls, file_path):
        file_path = Path(file_path)
        with file_path.open('r') as file:
            file_dict = safe_load(file)
        new_instance = cls(**file_dict)
        return new_instance


import os
import tempfile
from tests.test_base import TestBase
from capanno_utils.repo_config import config
from capanno_utils.content_maps import make_tools_map, make_script_maps


class TestToolMaps(TestBase):

    def test_make_tools_map(self):
        make_tools_map(tempfile.NamedTemporaryFile().name)
        return

    def test_make_tools_map_with_exists_status(self):
        make_tools_map(tempfile.NamedTemporaryFile().name, specify_exists=True)
        return

    def test_make_script_maps(self):
        make_script_maps(tempfile.NamedTemporaryFile().name)
        return
